import uuid
from datetime import datetime, UTC

from google.cloud import bigquery
from loguru import logger

from bb_integrations_lib.models.pipeline_structs import BolExportResults
from bb_integrations_lib.protocols.pipelines import Step


class UpdateExportedOrdersTableStep(Step[BolExportResults, BolExportResults, None]):
    def __init__(
            self,
            bigquery_client: bigquery.Client,
            exported_order_table_name: str,
            exported_order_errors_table_name: str,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.bigquery_client = bigquery_client
        self.exported_order_table_name = exported_order_table_name
        self.exported_order_errors_table_name = exported_order_errors_table_name

    def describe(self) -> str:
        return "Update the GCP bigquery table to include the newly-exported orders"

    async def execute(self, results: BolExportResults) -> BolExportResults:
        orders = results.orders
        unique_order_numbers_exported = set(x[results.order_number_key] for x in orders)
        now = datetime.now(UTC).isoformat().split("+")[0]
        if len(unique_order_numbers_exported) > 0:
            to_insert = [{
                "file_name": results.file_name,
                "date": now,
                "order_number": x
            } for x in unique_order_numbers_exported]
            logger.debug(f"Inserting {len(to_insert)} records for run {results.file_name}")
            gbq_errors = self.bigquery_client.insert_rows_json(self.exported_order_table_name, to_insert)
            if gbq_errors:
                logger.error(gbq_errors)
            else:
                logger.debug(f"Updated {self.exported_order_table_name}")
        else:
            logger.debug("No exported orders to insert.")

        errors_insert = [{
            "id": str(uuid.uuid4()),
            "export_id": results.file_name,
            "order_number": str(x["order_number"]),
            "error": x["error"],
            "date": now
        } for x in results.errors]
        if len(errors_insert) > 0:
            logger.debug(f"Inserting {len(errors_insert)} records for run {results.file_name}")
            gbq_errors = self.bigquery_client.insert_rows_json(self.exported_order_errors_table_name, errors_insert)
            if gbq_errors:
                logger.error(gbq_errors)
            else:
                logger.debug(f"Updated {self.exported_order_errors_table_name}")
        else:
            logger.debug(f"No errors to insert.")

        return results



import csv
from datetime import datetime, UTC
from typing import Optional, Iterable, Self

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.pipelines.steps.export_dataframe_to_rawdata_step import ExportDataFrameToRawDataStep
from bb_integrations_lib.pipelines.steps.exporting.bbd_export_readings_step import BBDExportReadingsStep
from bb_integrations_lib.pipelines.steps.exporting.sftp_export_file_step import SFTPExportFileStep
from bb_integrations_lib.pipelines.steps.processing.bbd_format_tank_readings_step import ParseTankReadingsStep
from bb_integrations_lib.pipelines.steps.processing.tank_reading_touchup_steps import TRTouchUpStep
from bb_integrations_lib.pipelines.steps.send_attached_in_rita_email_step import SendAttachedInRitaEmailStep
from bb_integrations_lib.protocols.pipelines import JobPipeline
from bb_integrations_lib.secrets import SecretProvider
from bb_integrations_lib.secrets.factory import APIFactory
from bb_integrations_lib.shared.model import ExportReadingsConfig, FileFormat, ExportReadingsMultiConfig
from loguru import logger


class BBDReadingExportPipeline(JobPipeline):
    @classmethod
    async def create(
            cls,
            config: ExportReadingsConfig,
            secret_provider: SecretProvider,
            rita_client: GravitateRitaAPI,
            bucket_name: str,
            config_name: str,
            touchup_step: Optional[TRTouchUpStep] = None,
            file_date: Optional[datetime] = None,
            send_reports: bool = True
    ):
        api_factory = APIFactory(secret_provider)

        # Create clients
        sd_client = await api_factory.sd(config.sd_credentials)
        ims_database = await api_factory.mongo_db(config.ims_credentials)

        ftp_client = None
        if config.ftp_credentials and config.ftp_directory:
            ftp_client = await api_factory.ftp(config.ftp_credentials)

        file_date = file_date or datetime.now(UTC)
        file_name = f"{config.file_base_name}_{file_date.strftime(config.file_name_date_format)}.csv"

        steps = [
            {
                "id": "1",
                "parent_id": None,
                "step": BBDExportReadingsStep(
                    rita_client=rita_client,
                    sd_client=sd_client,
                    ims_database=ims_database,
                    reading_query=config.reading_query,
                    timezone=config.reading_reported_timezone,
                    window_mode=config.window_mode,
                    hours_back=config.hours_back,
                )
            },
            {
                "id": "2",
                "parent_id": "1",
                "step": ParseTankReadingsStep(
                    file_format=config.file_format,
                    timezone=config.reading_reported_timezone,
                    include_water_level=config.include_water_level,
                    disconnected_column=config.disconnected_column,
                    disconnected_only=config.disconnected_only,
                    disconnected_hours_threshold=config.disconnected_hours_threshold,
                )
            },
        ]

        if touchup_step is not None:
            steps.append({
                "id": "touchup",
                "parent_id": "2",
                "step": touchup_step,
            })
            final_build_step = "touchup"
        else:
            final_build_step = "2"

        pd_export_function = "write_csv" if config.use_polars else "to_csv"
        if config.use_polars:
            pd_kwargs: dict = {"include_header": True}
            if config.file_format == FileFormat.reduced:
                pd_kwargs["quote_style"] = "never"
        else:
            pd_kwargs = {"header": True, "index": False}
            if config.file_format == FileFormat.reduced:
                pd_kwargs["quoting"] = csv.QUOTE_NONE
                pd_kwargs["escapechar"] = "\\"

        steps.append({
            "id": "3",
            "parent_id": final_build_step,
            "step": ExportDataFrameToRawDataStep(
                pandas_export_function=pd_export_function,
                pandas_export_kwargs=pd_kwargs,
                file_name=file_name,
            )
        })

        if config.ftp_directory is not None and ftp_client is not None:
            steps.append({
                "id": "4",
                "parent_id": "3",
                "step": SFTPExportFileStep(
                    ftp_client=ftp_client,
                    ftp_destination_dir=config.ftp_directory,
                )
            })

        if config.email_addresses is not None:
            steps.append({
                "id": "send_email",
                "parent_id": "3",
                "step": SendAttachedInRitaEmailStep(
                    rita_client=rita_client,
                    to=config.email_addresses,
                    html_content="Tank Reading Export",
                    subject=f"Gravitate Tank Reading Export - {datetime.now().isoformat()}",
                    timeout=30.0,
                )
            })

        config = await rita_client.get_config_by_name(bucket_path="/" + bucket_name, config_name=config_name)
        config_id = config[config_name].config_id
        return cls(
            steps,
            rita_client=rita_client,
            pipeline_name=f"Import tank readings - {config_name}",
            pipeline_config_id=config_id,
            secret_provider=secret_provider,
            send_reports=send_reports
        )

class MultiExportTankReadingsPipeline:
    def __init__(self, pipelines: Iterable[BBDReadingExportPipeline]):
        self.pipelines = pipelines

    @classmethod
    async def create(cls, config_id: str, rita_client: GravitateRitaAPI,
                     secret_provider: SecretProvider, send_reports: bool = True) -> Self:
        main_config = await rita_client.get_config_by_id(config_id)
        [bucket] = [b for b in await rita_client.get_all_buckets() if b["_id"] == main_config.owning_bucket_id]

        multi_export_config = ExportReadingsMultiConfig.model_validate(main_config.config)
        pipelines = []
        for export_config in multi_export_config.configs:
            for config_name in export_config.config_names:
                logger.info(f"Creating pipeline for config {config_name}")
                pipelines.append(await BBDReadingExportPipeline.create(
                    config=export_config,
                    secret_provider=secret_provider,
                    rita_client=rita_client,
                    bucket_name=bucket["name"],
                    config_name=config_name,
                    send_reports=send_reports,
                ))
        return cls(pipelines)

    async def run(self) -> None:
        if not self.pipelines:
            raise Exception("No pipelines to run - check config and try again")
        for pipeline in self.pipelines:
            try:
                logger.info(f"Starting pipeline '{pipeline.pipeline_name}'")
                await pipeline.execute()
                logger.info(f"Finished pipeline '{pipeline.pipeline_name}'")
                logger.info("-" * 40)
            except Exception as e:
                logger.exception(f"Failed to run pipeline '{pipeline.pipeline_name}'")



from enum import Enum


class ProbeEventType(str, Enum):
    create = "create"
    update = "update"
    delete = "delete"


class CrossroadsEntityType(str, Enum):
    """Supported entities for crossroads integrations."""
    # For watching orders in a generic manner (i.e. report when created). Planned for use with Telapoint integration
    order = "order"

    # For watching orders specifically for the carrier order integration (which will only be kicked off when an order is assigned to a carrier)
    carrier_order = "carrier order integration"


class CrossroadsTaskStatus(str, Enum):
    created = "created"
    started = "started"
    succeeded = "succeeded"
    failed = "failed"


class ConnectorAction(str, Enum):
    carrier_integration_create_order = "carrier integration - create order for carrier"
    carrier_integration_update_order_status = "carrier integration - update order status for customer"



from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field


class HierarchyKey(BaseModel):
    id: str | None = None
    name: str | None = None
    source_system: str | None = None
    source_id: str | None = None


class SiteKey(BaseModel):
    id: str | None = None
    number: str | None = None
    source_system: str | None = None
    source_id: str | None = None


class OrderKey(BaseModel):
    order_number: int | None
    order_id: str | None = None
    order_reference_number: str | None = None


class DeliveryWindow(BaseModel):
    start: datetime
    end: datetime
    timezone: str = "UTC"


class LoadsCreateRequest(BaseModel):
    terminal: HierarchyKey | None = None
    product: HierarchyKey | None = None
    supplier: HierarchyKey | None = None
    price_type: str | None = None


class DropCreateRequest(BaseModel):
    site: SiteKey
    tank_id: str | None = None
    product: HierarchyKey
    volume: int
    loads: list[LoadsCreateRequest] = []


class OrderCreateRequest(BaseModel):
    reference_order_number: str
    supply_owner: HierarchyKey | None = None
    sourcing_strategy: Literal["Specific Supply", "Tank Supply Default", "Use Best", "Manual"] = "Specific Supply"
    manual_supply_fallback: bool = True
    allow_alternate_products: bool = False
    delivery_window: DeliveryWindow
    fit_to_trailer: bool = False

    note: str | None = None
    drops: list[DropCreateRequest]
    extra_data: dict


class OrderStatusUpdateRequest(BaseModel):
    order_id: str
    order_number: str | None = None
    status: str
    location_id: str | None = None
    eta: datetime | None = None
    actual: datetime | None = None

class DropDetail(BaseModel):
    product_id: str
    quantity: float
    tank_id: int | None = None
    pre_drop_volume: float | None = None
    pre_drop_inches: float | None = None
    pre_drop_time: datetime | None = None
    post_drop_volume: float | None = None
    post_drop_inches: float | None = None
    post_drop_time: datetime | None = None

class SaveDropDetailsRequest(BaseModel):
    order_id: str
    location_id: str
    mode: Literal["replace", "append"]
    details: list[DropDetail]

class BOLDetail(BaseModel):
    supplier_id: str
    product_id: str
    contract: str | None = None
    price_type: str | None = None
    net_volume: float
    gross_volume: float

class SaveBOLDetailsRequest(BaseModel):
    order_id: str
    bol_number: str
    terminal_id: str
    bol_date: datetime
    details: list[BOLDetail]

class OrderUpdateRequest(BaseModel):
    order: OrderKey = Field(...)
    sourcing_strategy: Literal["Specific Supply", "Tank Supply Default", "Use Best", "Manual"] = Field(default="Use Best")
    delivery_window: DeliveryWindow | None = Field(default=None, description="If provided, replaces the existing delivery window.")
    note: str | None = Field(default=None, description="If provided, replaces the existing note.")
    drops: list[DropCreateRequest] = Field(default=[], description="If provided, replaces all existing drops.")
    extra_data: dict = Field(default={}, description="Merge this extra data with the existing extra data. New keys will be added, existing keys will be updated.")

class SendToCarrierRequest(BaseModel):
    order_ids: list[str]
    force: bool = False
    instant: bool = False
    actual_end_date: datetime | None = None
    send_email: bool = True


import warnings

from bb_integrations_lib.provider.aws.s3.client import S3Client

class AWSIntegration(S3Client):
    def __init__(self, *args, **kwargs):
        warnings.warn("Use provider.aws.s3.client.S3Client", DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)



from datetime import datetime, UTC
from enum import Enum
from typing import Optional, Union, Literal, Annotated, Any

from pydantic import BaseModel, Field


class UnknownRunnableException(Exception):
    pass

WorkerTaskState = Literal["sent", "received", "working", "failed", "completed"]

class WorkerRequest(BaseModel):
    """A request for a specific runnable to be executed on an available worker."""
    runnable_name: str
    runnable_kwargs: dict[str, Any] = {}
    tenant_name: str
    originator: Literal["backend", "probe", "other"] = "other"

class WorkerContentFile(BaseModel):
    """
    A file included in a worker's response. Supports binary (non-text) files by automatically encoding/decoding from
    base64 when serialized.
    """
    file_name: str
    mime_type: str | None = None
    content: bytes | None = None

    class Config:
        ser_json_bytes = "base64"
        val_json_bytes = "base64"

class WorkerResponseStatus(str, Enum):
    success = "success"
    error = "error"

class WorkerSuccessResponse(BaseModel):
    """A response from a worker indicating success and providing results."""
    # This is a pretty weird field definition. Basically, to support using this field as a discriminator for the
    # WorkerResponse union type, we need it to only ever be status="success", (or "error", in the error class).
    # These type annotations mean it is a WorkerResponseStatus that can only ever be .success, and that it cannot be set
    # in the init method (init=False) or anywhere else (init_var=True).
    # You can't use a frozen field here because it won't play nice with Beanie serialization.
    status: Annotated[Literal[WorkerResponseStatus.success], Field(init_var=True, init=False)] = WorkerResponseStatus.success
    runnable_name: str
    content: str | None = None
    content_file: WorkerContentFile | None = None
    extra_data: dict = {}

class WorkerErrorResponse(BaseModel):
    """A response from a worker indicating an error occurred and providing error details."""
    status: Annotated[Literal[WorkerResponseStatus.error], Field(init_var=True, init=False)] = WorkerResponseStatus.error
    runnable_name: str
    exception_type_name: str
    error_message: str
    reference_code: Optional[str] = None

WorkerResponse = Annotated[Union[WorkerSuccessResponse, WorkerErrorResponse], Field(discriminator="status")]


class WorkerUpdate(BaseModel):
    """Provides status updates to the requester for backgrounded tasks."""
    tenant_name: str
    originator: str
    sent_at: datetime = Field(default_factory=lambda: datetime.now(tz=UTC))
    state: WorkerTaskState
    response: Optional[WorkerResponse] = None

class WorkerTask(BaseModel):
    correlation_id: str
    created_at: datetime
    updated_at: datetime
    mode: Literal["immediate", "background"]
    state: WorkerTaskState
    request: WorkerRequest
    response: Optional[WorkerResponse] = None



from datetime import datetime, UTC
from typing import Dict

import pytz
from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.rita.config import MaxSync
from bb_integrations_lib.protocols.pipelines import Step
from dateutil.parser import parse


class GetLatestSyncDate(Step):
    def __init__(
            self,
            rita_client: GravitateRitaAPI,
            config_id: str | None = None,
            test_override: str | None = None,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.rita_client = rita_client
        self.config_id = config_id
        self.test_override = test_override

    def describe(self) -> str:
        return "Get Latest Sync Date"

    async def execute(self, i: str) -> datetime:
        return await self.get_last_sync_date()

    async def get_last_sync_date(self) -> datetime:
        if self.test_override:
            return parse(self.test_override).replace(tzinfo=pytz.UTC)
        if not self.config_id or self.config_id is None:
            return datetime.now(UTC)
        max_sync: MaxSync  = await self.rita_client.get_config_max_sync(config_id=self.config_id)
        dt = max_sync.max_sync_date.replace(tzinfo=pytz.UTC)
        return dt



from typing import Optional, Union, List, Dict
from collections.abc import Iterable
from loguru import logger

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.mappers.prices.model import PriceMappings, Group, PricingIntegrationConfig, \
    IntegrationMappingConfig, IntegrationType
from bb_integrations_lib.models.rita.mapping import Map
from bb_integrations_lib.provider.sqlserver.client import SQLServerClient
from bb_integrations_lib.util.cache.custom_ttl_cache import CustomTTLCache
from bb_integrations_lib.util.cache.protocol import CustomTTLCacheProtocol


class PriceMapper:
    """Mapper instance to get all mapping data for needed for prices.
    Limited to:
        - Product, Location, Supplier and Publisher mappings
    """
    def __init__(
            self,
            config: PricingIntegrationConfig,
            sql_client: SQLServerClient | None = None,
            rita_client: GravitateRitaAPI | None = None,
            ttl: int = 3600,
            debug_mode: bool = False,
            source_system: Optional[str] = None
    ):
        self.ttl = ttl
        self.config = config
        self.sql_client = sql_client
        self.rita_client = rita_client
        self.products_config: IntegrationMappingConfig = self.config.entity_config.get("products").mapping_integration
        self.locations_config: IntegrationMappingConfig = self.config.entity_config.get("locations").mapping_integration
        self.suppliers_config: IntegrationMappingConfig = self.config.entity_config.get("suppliers").mapping_integration
        self.publishers_config: IntegrationMappingConfig = self.config.entity_config.get("publishers").mapping_integration
        self.products_sql_query = self.products_config.query
        self.locations_sql_query = self.locations_config.query
        self.suppliers_sql_query = self.suppliers_config.query
        self.cache: CustomTTLCacheProtocol = CustomTTLCache(verbose=debug_mode)
        self.debug_mode = debug_mode
        self.source_system = source_system

        if debug_mode:
            logger.debug("DEBUG MODE ON")

    @property
    def ttl_cache(self):
        return self.cache

    async def _get_cached_data(self, cache_key: str, query_func) -> Iterable:
        """Generic method to retrieve cached data or fetch it using the provided query function"""

        @self.cache.ttl_cache(seconds=self.ttl, cache_key=cache_key)
        async def _get_data() -> Iterable:
            return await query_func() or {}
        return await _get_data()


    async def get_entity_mappings(self, key: str, mapping_type: Optional[str] = None, query: Optional[str] = None) -> Iterable:
        entity_config = self.config.entity_config.get(key)
        integration_type = entity_config.mapping_integration.type if entity_config and entity_config.mapping_integration else None

        async def query_func():
            if integration_type == IntegrationType.sql:
                if not self.sql_client:
                    raise ValueError(f"SQL client required for entity '{key}' but not provided")
                return self.sql_client.get_mappings(query)
            elif integration_type == IntegrationType.rita:
                if not self.rita_client:
                    raise ValueError(f"RITA client required for entity '{key}' but not provided")
                return await self.rita_client.get_mappings(source_system=self.source_system, mapping_type=mapping_type)
            return {}
        return await self._get_cached_data(key, query_func)


    async def get_mappings(
            self,
            product_key: Optional[str] = "products",
            location_key: Optional[str] = "locations",
            supplier_key: Optional[str] = "suppliers",
            price_publisher_key: Optional[str] = "publishers"
    ) -> PriceMappings:
        return PriceMappings(
            product_mappings=await self.get_entity_mappings(product_key, mapping_type="product", query=self.products_sql_query),
            location_mappings=await self.get_entity_mappings(location_key, mapping_type="location", query=self.locations_sql_query),
            supplier_mappings=await self.get_entity_mappings(supplier_key, mapping_type="supplier", query=self.suppliers_sql_query),
            price_publishers=await self.get_entity_mappings(price_publisher_key, mapping_type="other")
        )

    @classmethod
    def group_rows(
            cls,
            rows: Union[List[Map], List[Dict]],
            external_id_field: str,
            gravitate_id_field: str,
            name_field: Optional[str] = None,
            is_rita: Optional[bool] = False,
    ) -> Dict[
        str, Group]:
        grouped = {}
        if is_rita:
            rows: List[Map]
            for row in rows:
                row_id = row.source_id
                children = [r.gravitate_id for r in row.children]
                if row_id not in grouped:
                    grouped[row_id] = Group(
                        name=row.gravitate_name,
                        ids=children,
                        length=len(children),
                        extra_data=row
                    )
            return grouped
        else:
            rows: List[Dict]
            for row in rows:
                row_id = row[external_id_field]
                row_guid = row[gravitate_id_field]
                row_name = row.get(name_field) if name_field else None
                if row_id not in grouped:
                    grouped[str(row_id)] = Group(
                        name=row_name,
                        ids=[row_guid],
                        length=1,
                        extra_data=row
                    )
                else:
                    group = grouped[row_id]
                    group.ids.append(row_guid)
                    group.length += 1
        return grouped





import json
import math
from enum import Enum
from typing import Dict, Iterable, List, Optional

import pandas as pd
import pytz
from dateutil.parser import parse
from loguru import logger
import traceback

from bb_integrations_lib.models.rita.mapping import MappingType
from bb_integrations_lib.shared.model import RawData, FileConfigRawData, MappingMode
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.models.rita.config import FileConfig, ConfigAction
from bb_integrations_lib.models.rita.issue import IssueBase, IssueCategory
from bb_integrations_lib.protocols.flat_file import TankReading, TankMonitorType, PriceRow, DriverCredential, \
    BulkSyncIntegrationDTO, PriceInstrumentDTO, PELookup, PriceTypeDTO, PeBulkSyncIntegration, \
    PriceMergeIntegrationDTO, PriceMergeValue, PePriceMergeIntegration, TankSales
from bb_integrations_lib.mappers.rita_mapper import RitaMapper, RitaAPIMappingProvider
from babel.numbers import parse_decimal
from zipfile import BadZipFile
from dateutil.tz import gettz

import warnings

tzmapping = {
    'EST': gettz("US/Eastern"),
    'EDT': gettz("US/Eastern"),
    'CST': gettz("US/Central"),
    'CDT': gettz("US/Central"),
    'MST': gettz("US/Mountain"),
    'MDT': gettz("US/Mountain"),
    'PST': gettz("US/Pacific"),
    'PDT': gettz("US/Pacific"),
}


class FileConfigParserStep(
    Step[FileConfigRawData, Iterable[TankReading] | Iterable[PriceRow] | Iterable[
        DriverCredential] | Iterable[PeBulkSyncIntegration] | Iterable[PePriceMergeIntegration] | Iterable[
                                TankSales], None]):
    def __init__(self, step_configuration: Dict[str, str]):
        super().__init__(step_configuration)
        self.rita_url = step_configuration["rita_url"]
        self.rita_psk = None
        self.client_id = step_configuration["client_id"]
        self.client_secret = step_configuration["client_secret"]
        self.rita_tenant = step_configuration["rita_tenant"]
        self.output_type = step_configuration.get('output_type', "TankReading")
        self.verbose = step_configuration.get("verbose", False)
        self.mapping_type = step_configuration.get("mapping_type", MappingMode.full)
        self.included_payload = step_configuration.get("included_payload", {})

        warnings.warn(
            "FileConfigParserStep is deprecated. Use FileConfigParserV2 + a custom parser instead.",
            DeprecationWarning, stacklevel=2
        )

    def describe(self) -> str:
        return f"Parse Flat File into {self.output_type}"

    async def execute(self, i: FileConfigRawData) -> Iterable[TankReading] | Iterable[PriceRow] | Iterable[
        DriverCredential] | Iterable[PeBulkSyncIntegration] | Iterable[PePriceMergeIntegration | Iterable[TankSales]]:
        assert isinstance(i, FileConfigRawData)
        logger.info(f"Beginning data extract for file {i.file_name}")
        logger.debug(
            f"File Config: {i.file_config.client_name}, Inbound: {i.file_config.inbound_directory}, Archive: {i.file_config.archive_directory}")

        return list(await self.parse_raw_data(i))

    def parse_tank_and_site_ids(self, records: List[Dict], rd: FileConfigRawData, mapper: RitaMapper) -> Dict:
        """Only to be used by TankReading or TankSales"""
        translated_records = {}
        translation_failures = []
        mapping_failures = []
        for row in records:
            try:
                translated = self.translate_row(rd.file_config, row)
            except Exception as e:
                if self.verbose:
                    logger.warning(f"Translation failed for record {row}: {e}")
                translation_failures.append(row)
                continue
            try:
                if 'volume' in translated:
                    if translated['volume'] == 'nan':
                        if self.verbose:
                            logger.warning(f"Skipped record {row} due to NaN volume. Translated={translated}")
                        continue
                    if float(translated["volume"]) < 0:
                        if self.verbose:
                            logger.warning(f"Skipped record {row} due to Negative volume. Translated={translated}")
                        continue
                if 'tank_id' in translated:
                    if translated['tank_id'] == 'nan':
                        if self.verbose:
                            logger.warning(f"Skipped record {row} due to NaN tank. Translated={translated}")
                        continue
                if 'sales' in translated:
                    if translated['sales'] == 'nan':
                        if self.verbose:
                            logger.warning(f"Skipped record {row} due to NaN sales. Translated={translated}")
                        continue
                if self.mapping_type == MappingMode.skip:
                    key = f"{translated["site_id"]}_{translated["tank_id"]}"
                    translated_records[key] = translated
                elif self.mapping_type == MappingMode.partial or self.mapping_type == MappingMode.full:
                    try:
                        site_id = translated['site_id']
                        tank_id = translated["tank_id"]
                        mapped_site_ids = mapper.get_gravitate_parent_ids(site_id, MappingType.site)
                        mapped_tank_ids = mapper.get_gravitate_child_ids(site_id, tank_id.strip(), MappingType.tank)
                        for site_id in mapped_site_ids:
                            for tank_id in mapped_tank_ids:
                                key = f"{site_id}_{tank_id}"
                                translated["site_id"] = site_id
                                translated["tank_id"] = tank_id
                                translated_records[key] = {**translated}
                    except (KeyError, ValueError) as e:
                        if self.mapping_type == MappingMode.partial:
                            key = f"{translated["site_id"]}_{translated["tank_id"]}"
                            translated_records[key] = translated
                        else:
                            raise e
                else:
                    raise ValueError(f"MappingType {self.mapping_type} not supported")


            except (KeyError, ValueError) as e:
                if self.verbose:
                    logger.warning(f"Skipped record {row} due to error: {e}")
                mapping_failures.append(translated)
        self.handle_issues(translation_failures, mapping_failures, rd, output_type=self.output_type)
        return translated_records

    def parse_price_row(self, records: List[Dict], rd: FileConfigRawData, mapper: Optional[RitaMapper] = None) -> \
            Iterable[PriceRow]:
        translation_failures = []
        mapping_failures = []
        for row in records:
            try:
                translated = self.translate_row(rd.file_config, row)
                yield PriceRow(
                    effective_from=translated["effective_from"],
                    effective_to=translated["effective_to"],
                    price=float(translated["price"]),
                    price_type=translated["price_type"],
                    contract=translated.get("contract", None),
                    timezone=translated.get("timezone", None),
                    terminal_id=translated.get("terminal_id", None),
                    terminal_source_id=translated.get("terminal_source_id", None),
                    terminal_source_system_id=translated.get("terminal_source_system_id", None),
                    terminal=translated.get("terminal", None),
                    product_id=translated.get("product_id", None),
                    product_source_id=translated.get("product_source_id", None),
                    product_source_system_id=translated.get("product_source_system_id", None),
                    product=translated.get("product", None),
                    supplier_id=translated.get("supplier_id", None),
                    supplier_source_id=translated.get("supplier_source_id", None),
                    supplier_source_system_id=translated.get("supplier_source_system_id", None),
                    counterparty_source_id=translated.get("counterparty_source_id", None),
                    counterparty_source_system_id=translated.get("counterparty_source_system_id", None),
                    supplier=translated.get("supplier", None),
                    enabled=bool(translated.get("enabled", None)),
                    disabled_until=translated.get("disabled_until", None),
                    expire=translated.get("expire", None),
                    min_quantity=translated.get("min_quantity", None),
                    max_quantity=translated.get("max_quantity", None),
                    curve_id=translated.get("curve_id", None),
                    row=translated.get("row", None),
                )
            except(KeyError, ValueError) as e:
                translation_failures.append(row)
                if self.verbose:
                    logger.warning(f"Skipped record {row} due to error: {e}")
        self.handle_issues(translation_failures, mapping_failures, rd, output_type=self.output_type)

    def parse_tank_reading_rows(self, records: List[Dict], rd: FileConfigRawData,
                                mapper: Optional[RitaMapper] = None) -> Iterable[TankReading]:
        translated_records = self.parse_tank_and_site_ids(records, rd, mapper)

        for translated_record in translated_records.values():
            if translated_record.get("reading_time") is None:
                logger.warning(f"Skipped record {translated_record} due missing date")
                continue
            try:
                date_parsed = parse(translated_record.get("reading_time"), tzinfos=tzmapping).isoformat()
            except Exception as parse_error:
                logger.warning(f"Skipped record {translated_record} due to date parsing error: {parse_error}")
                continue
            try:
                yield TankReading(
                    store=translated_record.get("site_id"),
                    date=date_parsed,
                    monitor_type=TankMonitorType.bbd,
                    timezone=translated_record.get("timezone"),
                    volume=translated_record.get("volume"),
                    tank=translated_record.get("tank_id"),
                    payload=self.included_payload

                )
            except ValueError as e:
                # ValueError here probably means a volume has a , in the number or is a string. Parse
                # it with Babel instead.
                yield TankReading(
                    store=translated_record.get("site_id"),
                    date=date_parsed,
                    monitor_type=TankMonitorType.bbd,
                    timezone=translated_record.get("timezone"),
                    volume=float(parse_decimal(translated_record.get("volume"), locale="en_US")),
                    tank=translated_record.get("tank_id"),
                    payload=self.included_payload
                )
            except Exception as ee:
                logger.warning(f"Skipped record {translated_record} due to error: {ee}")
                continue

    def parse_tank_sales_rows(self, records: List[Dict], rd: FileConfigRawData, mapper: Optional[RitaMapper] = None) -> \
            Iterable[TankSales]:
        translated_records = self.parse_tank_and_site_ids(records, rd, mapper)
        ll = []
        errors = []
        for key in translated_records:
            site_id = translated_records[key].get("site_id", "")
            tank_id = translated_records[key].get("tank_id", "")
            try:
                tz = translated_records.get(key, {}).get("timezone")
                ll.append(TankSales(
                    store_number=site_id,
                    tank_id=tank_id,
                    sales=float(translated_records.get(key)['sales']),
                    date=parse(translated_records.get(key)['date'], tzinfos=tzmapping).replace(
                        tzinfo=(pytz.timezone(tz)), hour=0, minute=0, second=0, microsecond=0).isoformat(),
                ))
            except ValueError as e:
                errors.append({
                    "error": f"ValueError: {str(e)}",
                    "row": f"record containing site: {site_id} and tank: {tank_id}",
                })
                if self.verbose:
                    logger.warning(
                        f"Skipped containing site: {site_id} and tank: {tank_id} due to error: {str(e)}")
                continue
            except KeyError as ke:
                errors.append({
                    "error": f"KeyError: {str(ke)}",
                    "row": f"record containing site: {site_id} and tank: {tank_id}",
                })
                if self.verbose:
                    logger.warning(
                        f"Skipped record containing site: {site_id} and tank: {tank_id} due to error: {str(ke)}")
                continue
            except TypeError as tpe:
                errors.append({
                    "error": f"TypeError: {str(tpe)}",
                    "row": f"record containing site: {site_id} and tank: {tank_id}",
                })
                if self.verbose:
                    logger.warning(
                        f"Skipped record containing site: {site_id} and tank: {tank_id} due to error: {str(tpe)}")
                continue
            except Exception as ee:
                errors.append({
                    "error": f"Unknown error: {str(ee)}",
                    "row": f"record containing site: {site_id} and tank: {tank_id}",
                })
                if self.verbose:
                    logger.warning(
                        f"Skipped record containing site: {site_id} and tank: {tank_id} due to error: {str(ee)}")
                continue
        return ll

    def parse_credential_rows(self, records: List[Dict], rd: FileConfigRawData, mapper: Optional[RitaMapper] = None) -> \
            Iterable[DriverCredential]:
        translation_failures = []
        mapping_failures = []
        for row in records:
            try:
                translated = self.translate_row(rd.file_config, row)
                credential_ids = mapper.get_gravitate_parent_ids(source_id=translated['credential_id'],
                                                                 mapping_type=MappingType.credential)
                driver_ids = mapper.get_gravitate_parent_ids(source_id=translated['driver_id'],
                                                             mapping_type=MappingType.driver)
                expiration_date = translated.get('expiration_date')
                if expiration_date in ['nan']:
                    if self.verbose:
                        logger.warning(f"Skipped record due to NaN expiration date")
                    continue
                for credential_id in credential_ids:
                    for driver_id in driver_ids:
                        yield DriverCredential(
                            driver_id=driver_id,
                            credential_id=credential_id,
                            certification_date=translated.get('certification_date'),
                            expiration_date=expiration_date
                        )
            except (KeyError, ValueError) as e:
                translation_failures.append(row)
                if self.verbose:
                    logger.warning(f"Skipped record {row} due to error: {e}")
                continue
            self.handle_issues(translation_failures, mapping_failures, rd, output_type=self.output_type)

    def parse_price_sync_rows(self, records: List[Dict], rd: FileConfigRawData, mapper: Optional[RitaMapper] = None) -> \
            Iterable[PeBulkSyncIntegration]:
        translation_failures = []
        mapping_failures = []
        dtos: List[PriceInstrumentDTO] = []
        price_publisher_name = None
        posting_type = "Posting"
        source_system_id = None
        for row in records:
            try:
                row_is_rvp = FileConfigParserStep.is_rvp(row.get('RVP', '0.0'))
                translated = self.translate_row(rd.file_config, row)
                price_publisher_name = translated['price_publisher']
                configuration = translated['configuration']
                supplier_key = translated['supplier_key']
                location_key = translated['location_key']
                product_key = translated['product_key']
                posting_type = translated.get('posting_type', "Posting")
                location_name = translated.get('location_name', None)
                product_name = translated.get('product_name', None)
                supplier_name = translated.get('supplier_name', None)
                if row_is_rvp and not "RVP" in product_key:
                    product_key = FileConfigParserStep.format_rvp_product(product_key, row['RVP'])
                if '.' in product_key and not row_is_rvp:
                    product_key = product_key.split('.')[0]
                source_system_id = translated['source_system_id']

                if self.mapping_type == MappingMode.full:
                    supplier_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                       source_child_id=supplier_key,
                                                                       mapping_type=MappingType.counterparty)
                    location_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                       source_child_id=location_key,
                                                                       mapping_type=MappingType.terminal)
                    product_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                      source_child_id=product_key,
                                                                      mapping_type=MappingType.product)
                elif self.mapping_type == MappingMode.skip:
                    supplier_source_id = supplier_key
                    location_source_id = location_key
                    product_source_id = product_key
                else:
                    raise ValueError(f"Unsupported mapping type: {self.mapping_type}")
                if product_name is not None and location_name is not None and supplier_name is not None:
                    price_instrument_name = f"{product_name} @ {location_name} - {supplier_name}"
                else:
                    price_instrument_name = f"{product_source_id} @ {location_source_id} - {supplier_source_id}"
                price_instrument_source_string_id = f"{price_publisher_name} - {price_instrument_name}"
                dtos.append(PriceInstrumentDTO(
                    Name=price_instrument_name,
                    Abbreviation=price_instrument_name,
                    SourceIdString=price_instrument_source_string_id,
                    ProductLookup=PELookup(
                        SourceIdString=product_source_id,
                        SourceSystemId=int(source_system_id)
                    ),
                    LocationLookup=PELookup(
                        SourceIdString=location_source_id,
                        SourceSystemId=int(source_system_id)
                    ),
                    CounterPartyLookup=PELookup(
                        SourceIdString=supplier_source_id,
                        SourceSystemId=int(source_system_id)
                    )
                ))
            except (KeyError, ValueError) as e:
                mapping_failures.append(row)
                logger.warning(f"Skipped record {row} due to Key Error or Value Error: {e}")

                continue
            except Exception as uh:
                translation_failures.append(row)
                if self.verbose:
                    logger.warning(f"Skipped record {row} due to unhandled exception: {uh.args}")
                if irc := self.pipeline_context.issue_report_config:
                    if len(translation_failures) > 0:
                        key = f"{irc.key_base}_{rd.file_config.config_id}_translation_error"
                        self.pipeline_context.issues.append(IssueBase(
                            key=key,
                            config_id=rd.file_config.config_id,
                            name=f"Translation error - {rd.file_config.client_name}",
                            category=IssueCategory.TANK_READING,
                            problem_short=f"{rd.file_name}: At least 1 row failed to translate",
                            problem_long=json.dumps(translation_failures)
                        ))
                continue

        if dtos and price_publisher_name and source_system_id:
            try:
                yield PeBulkSyncIntegration(
                    IntegrationDtos=[
                        BulkSyncIntegrationDTO(
                            Name=price_publisher_name,
                            Abbreviation=price_publisher_name,
                            SourceIdString=price_publisher_name,
                            PriceInstrumentDTOs=dtos,
                            PriceTypeDTOs=[
                                PriceTypeDTO(
                                    PriceTypeMeaning=posting_type
                                )
                            ]
                        )
                    ],
                    SourceSystemId=int(source_system_id)
                )

            except Exception as ex:
                logger.warning(f"Unexpected error when yielding integration: {ex}")
                if irc := self.pipeline_context.issue_report_config:
                    key = f"{irc.key_base}_{rd.file_config.config_id}_parser_error"
                    self.pipeline_context.issues.append(IssueBase(
                        key=key,
                        config_id=rd.file_config.config_id,
                        name=f"Parser error - {rd.file_config.client_name}",
                        category=IssueCategory.REFERENCE_DATA,
                        problem_short=f"{rd.file_name}: Unexpected exception",
                        problem_long=json.dumps({
                            "exception_type": type(ex).__name__,
                            "args": list(ex.args) if ex.args else []
                        })
                    ))
        else:
            logger.warning("No valid records were processed to create an integration")

    def parse_price_merge_rows(self, records: List[Dict], rd: FileConfigRawData, mapper: Optional[RitaMapper] = None) -> \
            Iterable[PePriceMergeIntegration]:
        translation_failures = []
        mapping_failures = []
        dtos: List[PriceMergeIntegrationDTO] = []
        price_publisher_name = None
        posting_type = "Posting"  # Default value
        source_system_id = None
        for row in records:
            try:
                row_is_rvp = FileConfigParserStep.is_rvp(row.get('RVP', '0.0'))
                translated = self.translate_row(rd.file_config, row)
                price_publisher_name = translated['price_publisher']
                configuration = translated['configuration']
                supplier_key = translated['supplier_key']
                location_key = translated['location_key']
                product_key = translated['product_key']
                location_name = translated.get('location_name', None)
                product_name = translated.get('product_name', None)
                supplier_name = translated.get('supplier_name', None)
                price_factor = translated.get('price_factor')
                if row_is_rvp and not "RVP" in product_key:
                    product_key = FileConfigParserStep.format_rvp_product(product_key, row['RVP'])
                if '.' in product_key and not row_is_rvp:
                    product_key = product_key.split('.')[0]
                source_system_id = translated['source_system_id']
                posting_type = translated.get('posting_type', "Posting")
                date_str = translated['date']
                price = translated['price']
                if price_factor is not None:
                    price = float(price) / int(price_factor)

                if self.mapping_type == MappingMode.full:
                    supplier_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                       source_child_id=supplier_key,
                                                                       mapping_type=MappingType.counterparty)
                    location_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                       source_child_id=location_key,
                                                                       mapping_type=MappingType.terminal)
                    product_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                      source_child_id=product_key,
                                                                      mapping_type=MappingType.product)
                elif self.mapping_type == MappingMode.skip:
                    supplier_source_id = supplier_key
                    location_source_id = location_key
                    product_source_id = product_key
                else:
                    raise ValueError(f"Unsupported mapping type: {self.mapping_type}")
                if product_name is not None and location_name is not None and supplier_name is not None:
                    price_instrument_name = f"{product_name} @ {location_name} - {supplier_name}"
                else:
                    price_instrument_name = f"{product_source_id} @ {location_source_id} - {supplier_source_id}"
                price_instrument_source_string_id = f"{price_publisher_name} - {price_instrument_name}"
                dtos.append(PriceMergeIntegrationDTO(
                    PriceInstrumentLookup=PELookup(
                        SourceIdString=price_instrument_source_string_id,
                        SourceSystemId=int(source_system_id)
                    ),
                    EffectiveFromDateTime=parse(date_str).isoformat(),
                    PriceValues=[
                        PriceMergeValue(
                            Value=float(price),
                        )
                    ]
                ))


            except (KeyError, ValueError) as e:
                mapping_failures.append(row)
                logger.warning(f"Skipped record {row} due to Key Error or Value Error: {e}")
                continue
            except Exception as uh:
                translation_failures.append(row)
                if self.verbose:
                    logger.warning(f"Skipped record {row} due to unhandled exception: {uh.args}")
                if irc := self.pipeline_context.issue_report_config:
                    if len(translation_failures) > 0:
                        key = f"{irc.key_base}_{rd.file_config.config_id}_translation_error"
                        self.pipeline_context.issues.append(IssueBase(
                            key=key,
                            config_id=rd.file_config.config_id,
                            name=f"Translation error - {rd.file_config.client_name}",
                            category=IssueCategory.TANK_READING,
                            problem_short=f"{rd.file_name}: At least 1 row failed to translate",
                            problem_long=json.dumps(translation_failures)
                        ))
                continue
        if dtos and price_publisher_name and source_system_id:
            try:
                yield PePriceMergeIntegration(
                    IntegrationDtos=dtos,
                    SourceSystemId=int(source_system_id)
                )
            except Exception as ex:
                logger.warning(f"Unexpected error when yielding integration: {ex}")
                if irc := self.pipeline_context.issue_report_config:
                    key = f"{irc.key_base}_{rd.file_config.config_id}_parser_error"
                    self.pipeline_context.issues.append(IssueBase(
                        key=key,
                        config_id=rd.file_config.config_id,
                        name=f"Parser error - {rd.file_config.client_name}",
                        category=IssueCategory.PRICE,
                        problem_short=f"{rd.file_name}: Unexpected exception",
                        problem_long=json.dumps({
                            "exception_type": type(ex).__name__,
                            "args": list(ex.args) if ex.args else []
                        })
                    ))

    def handle_issues(self, translation_failures: List, mapping_failures: List, rd: FileConfigRawData,
                      output_type: Optional[str] = None):
        if not output_type or output_type is None:
            output_type = self.output_type
        if irc := self.pipeline_context.issue_report_config:
            if output_type == "TankReading":
                if len(translation_failures) > 0:
                    key = f"{irc.key_base}_{rd.file_config.config_id}_translation_error"
                    self.pipeline_context.issues.append(IssueBase(
                        key=key,
                        config_id=rd.file_config.config_id,
                        name=f"Translation error - {rd.file_config.client_name}",
                        category=IssueCategory.TANK_READING,
                        problem_short=f"{rd.file_name}: At least 1 row failed to translate",
                        problem_long=json.dumps(translation_failures)
                    ))

                if len(mapping_failures) > 0:
                    key = f"{irc.key_base}_{rd.file_config.config_id}_mapping_error"
                    self.pipeline_context.issues.append(IssueBase(
                        key=key,
                        config_id=rd.file_config.config_id,
                        name=f"Mapping error - {rd.file_config.client_name}",
                        category=IssueCategory.TANK_READING,
                        problem_short=f"{rd.file_config.client_name}: At least 1 row failed to map",
                        problem_long=json.dumps(mapping_failures)
                    ))

    async def parse_raw_data(self, rd: FileConfigRawData) -> Iterable[
        TankReading | PriceRow | DriverCredential | PeBulkSyncIntegration | PePriceMergeIntegration | TankSales]:
        try:
            records, raw_data, mapper = await self.get_records(rd)
            if self.output_type == "TankReading":
                return self.parse_tank_reading_rows(records, rd, mapper)
            if self.output_type == "TankSales":
                return self.parse_tank_sales_rows(records, rd, mapper)
            elif self.output_type == "PriceRow":
                return self.parse_price_row(records, rd, mapper)
            elif self.output_type == "Credential":
                return self.parse_credential_rows(records, rd, mapper)
            elif self.output_type == "PeStructureSync":
                return self.parse_price_sync_rows(records, rd, mapper)
            elif self.output_type == "PePriceMerge":
                return self.parse_price_merge_rows(records, rd, mapper)
            else:
                raise ValueError(
                    f"Invalid output_type: '{self.output_type}'. Only 'TankReading', 'PriceRow', 'Credential', 'PeStructureSync','PePriceMerge' or 'TankSales' are supported.")
        except Exception as e:
            logger.error(f"Error in parse_raw_data: {e}")
            if irc := self.pipeline_context.issue_report_config:
                key = f"{irc.key_base}_{rd.file_config.config_id}_parser_error"
                self.pipeline_context.issues.append(IssueBase(
                    key=key,
                    config_id=rd.file_config.config_id,
                    name=f"Parser error - {rd.file_config.client_name}",
                    category=IssueCategory.UNKNOWN,
                    problem_short=f"{rd.file_name}: Unexpected exception in parse_raw_data",
                    problem_long=json.dumps({
                        "exception_type": type(e).__name__,
                        "args": list(e.args) if e.args else []
                    })
                ))
            raise

    async def get_records(self, rd: RawData):  # Misstyped
        assert isinstance(rd, FileConfigRawData)
        try:
            if hasattr(rd.data, 'seek'):
                rd.data.seek(0)
            match rd.file_config.file_extension:
                case "csv1":
                    """This format will skip the top row of the file"""
                    temp_df = pd.read_csv(rd.data, index_col=False, dtype=str, skiprows=1)
                    temp_df = temp_df.rename(columns=lambda x: x.strip())
                    records = temp_df.to_dict(orient='records')
                case "csv":
                    temp_df = pd.read_csv(rd.data, index_col=False, dtype=str)
                    temp_df = temp_df.rename(columns=lambda x: x.strip())
                    records = temp_df.to_dict(orient="records")
                case "csv_headless":
                    """This format will treat the file as headless and load in generic column names 'col 1', 'col 2', etc."""
                    df = pd.read_csv(rd.data, index_col=False, dtype=str, header=None)
                    df.columns = [f"col {i + 1}" for i in range(df.shape[1])]
                    records = df.to_dict(orient='records')
                case "xls" | "xlsx":
                    """
                    Reads an excel and returns a dataframe
                    """
                    try:
                        temp_df = pd.read_excel(rd.data, engine="openpyxl", dtype=str)
                        temp_df = temp_df.rename(columns=lambda x: x.strip())
                        records = temp_df.to_dict(orient='records')
                    except (OSError, BadZipFile) as e:
                        # The file may be an old binary Excel file.
                        records = pd.read_excel(rd.data, engine="xlrd", dtype=str).to_dict(orient='records')
                case "html":
                    """
                    Reads an html and returns a dataframe
                    """
                    data = pd.read_html(rd.data, header=0)
                    merged = pd.concat(data)
                    records = merged.to_dict(orient="records")
                case _:
                    raise ValueError("The file_extension in the file config is not supported")

            rd.data_buffer_bkp = pd.DataFrame(records).to_csv(index=False)
            mapper = RitaMapper(
                provider=RitaAPIMappingProvider(
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    rita_tenant=self.rita_tenant,
                ),
                source_system=rd.file_config.source_system,
            )
            await mapper.load_mappings_async()

            return records, rd, mapper
        except Exception as e:
            if irc := self.pipeline_context.issue_report_config:
                key = f"{irc.key_base}_{rd.file_config.config_id}_failed_to_load"
                name = f"Failed to load file \"{rd.file_name}\" for client {rd.file_config.client_name}"
                self.pipeline_context.issues.append(IssueBase(
                    key=key,
                    config_id=rd.file_config.config_id,
                    name=name,
                    category=IssueCategory.TANK_READING,
                    problem_short=str(e),
                    problem_long=traceback.format_exc()
                ))
            raise e

    def translate_row(self, file_config: FileConfig, row: dict) -> Dict:
        output_row = {}
        for column in file_config.cols:
            if len(column.file_columns) == 0:
                output_row[column.column_name] = None
            elif column.action == ConfigAction.concat:
                concatenated = ""
                for entry in column.file_columns:
                    stripped_entry = entry.strip()
                    if stripped_entry in row:
                        value = row[stripped_entry]
                        if value is None or (isinstance(value, float) and math.isnan(value) or pd.isna(value)):
                            concatenated += ""
                        else:
                            concatenated += str(value)
                    else:
                        concatenated += str(entry)
                output_row[column.column_name] = concatenated
            elif column.action == ConfigAction.parse_date:
                output_row[column.column_name] = str(pd.to_datetime(row[column.file_columns[0]]))
            elif column.action == ConfigAction.add:
                output_row[column.column_name] = column.file_columns[0]
            elif column.action == ConfigAction.remove_leading_zeros:
                output_row[column.column_name] = self.strip_leading_zeroes(str(row[column.file_columns[0]]))
            elif column.action == ConfigAction.remove_trailing_zeros:
                output_row[column.column_name] = self.strip_trailing_zeroes(str(row[column.file_columns[0]]))
            else:
                output_row[column.column_name] = str(row[column.file_columns[0]])
        return output_row

    @staticmethod
    def strip_leading_zeroes(row: str):
        return row.lstrip('0')

    @staticmethod
    def strip_trailing_zeroes(row: str):
        return row.rstrip('0')

    @staticmethod
    def is_rvp(rvp: str) -> bool:
        try:
            return float(rvp) > 0.0
        except (ValueError, TypeError):
            return False

    @staticmethod
    def format_rvp_product(product_key: str, rvp: str) -> str:
        rvp_str = str(rvp)
        if product_key.endswith(rvp_str):
            product_key = product_key[:-len(rvp_str)]
        product_key = product_key.rstrip('.')
        return f"{product_key}{float(rvp_str)}"



from .credential_models import SDCredential, PECredential, RITACredential, IMAPCredential, AnyCredential, \
    BadSecretException, AbstractCredential, GoogleCredential, MongoDBCredential
from .credential_models import allowed_onepassword_models, onepassword_category_map
from .providers import SecretProvider, IntegrationSecretProvider



import functools
import traceback
from typing import Dict, Any, Callable, Optional
import json
from datetime import datetime


class ExceptionLogger:
    """
    A class that provides a decorator for catching and logging exceptions.
    """

    def __init__(self):
        """Initialize a new exception logger with an empty log."""
        self._exception_log: Dict[str, Any] = {}

    def catch_exceptions(self, identifier_param: Optional[str] = None):
        """
        Decorator that catches exceptions and logs them to this logger's dictionary.

        Args:
            identifier_param (str, optional): The parameter name to use as the identifier in the log.
                                             If None, will use function name + args as identifier.
        """

        def decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # Determine the identifier
                identifier = self._get_identifier(func, args, kwargs, identifier_param)

                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    # Log the exception
                    self._exception_log[identifier] = {
                        'timestamp': datetime.now().isoformat(),
                        'exception_type': type(e).__name__,
                        'exception_message': str(e),
                        'traceback': traceback.format_exc(),
                        'function': func.__name__,
                        'args': self._safe_repr(args),
                        'kwargs': self._safe_repr(kwargs)
                    }
                    # Re-raise the exception
                    raise

            return wrapper

        return decorator

    def _get_identifier(self, func: Callable, args: tuple, kwargs: dict, identifier_param: Optional[str]) -> str:
        """
        Get an identifier for the function call based on the provided parameters.

        Args:
            func: The function being called
            args: Positional arguments
            kwargs: Keyword arguments
            identifier_param: Parameter name to use as identifier

        Returns:
            str: An identifier for this function call
        """
        if identifier_param and identifier_param in kwargs:
            return str(kwargs[identifier_param])

        if identifier_param and args:
            # Try to find the position of the identifier parameter
            try:
                param_names = func.__code__.co_varnames[:func.__code__.co_argcount]
                if identifier_param in param_names:
                    idx = param_names.index(identifier_param)
                    if idx < len(args):
                        return str(args[idx])
            except (AttributeError, IndexError):
                pass

        # Generate a unique identifier using function name and a hash of arguments
        arg_str = self._safe_repr(args)
        kwarg_str = self._safe_repr(kwargs)
        return f"{func.__name__}_{hash(arg_str + kwarg_str)}_{datetime.now().timestamp()}"

    def _safe_repr(self, obj: Any) -> str:
        """
        Create a safe string representation of an object.

        Args:
            obj: The object to represent

        Returns:
            str: A string representation of the object
        """
        try:
            return str(obj)
        except Exception:
            return f"<unprintable object of type {type(obj).__name__}>"

    def get_log(self) -> Dict[str, Any]:
        """
        Get the current exception log.

        Returns:
            Dict[str, Any]: The exception log
        """
        return self._exception_log.copy()

    def clear_log(self) -> None:
        """Clear the exception log."""
        self._exception_log.clear()

    def get_formatted_log(self, indent: int = 2) -> str:
        """
        Get the exception log formatted as JSON.

        Args:
            indent (int): Number of spaces for JSON indentation

        Returns:
            str: Formatted JSON string
        """
        return json.dumps(self._exception_log, indent=indent)

    def get_exception_count(self) -> int:
        """
        Get the number of exceptions logged.

        Returns:
            int: Number of exceptions
        """
        return len(self._exception_log)

    def get_exceptions_by_type(self) -> Dict[str, int]:
        """
        Get a count of exceptions grouped by exception type.

        Returns:
            Dict[str, int]: Counts of each exception type
        """
        counts: Dict[str, int] = {}
        for exc_info in self._exception_log.values():
            exc_type = exc_info.get('exception_type', 'Unknown')
            counts[exc_type] = counts.get(exc_type, 0) + 1
        return counts





from typing import  Dict, List, override, AsyncGenerator

from loguru import logger

from bb_integrations_lib.models.rita.mapping import MappingType
from bb_integrations_lib.pipelines.parsers.price_engine.price_file_upload.shared import PriceSyncParser
from bb_integrations_lib.shared.model import MappingMode
from bb_integrations_lib.protocols.flat_file import PriceInstrumentDTO, PELookup, PeBulkSyncIntegration, \
    BulkSyncIntegrationDTO, PriceTypeDTO


class PricesSyncParser(PriceSyncParser):
    def __init__(self, tenant_name: str, source_system: str | None = None):
        super().__init__(tenant_name=tenant_name, source_system=source_system)

    def __repr__(self) -> str:
        return "Parse prices from rows to Sync Price Rows in Pricing Engine."

    @override
    async def parse(self, data: List[Dict], mapping_type: MappingMode | None = None) -> AsyncGenerator[
        PeBulkSyncIntegration, None]:
        mapper = await self.load_mapper()
        mapping_failures = []
        translation_failures = []
        dtos: List[PriceInstrumentDTO] = []
        for translated_row in data:
            try:
                row_is_rvp = PricesSyncParser.is_rvp(translated_row.get('RVP', '0.0'))
                price_publisher_name = translated_row['price_publisher']
                configuration = translated_row['configuration']
                supplier_key = translated_row['supplier_key']
                location_key = translated_row['location_key']
                product_key = translated_row['product_key']
                posting_type = translated_row.get('posting_type', "Posting")
                location_name = translated_row.get('location_name', None)
                product_name = translated_row.get('product_name', None)
                supplier_name = translated_row.get('supplier_name', None)
                if row_is_rvp and not "RVP" in product_key:
                    product_key = PricesSyncParser.format_rvp_product(product_key, translated_row['RVP'])
                source_system_id = translated_row['source_system_id']

                if mapping_type == MappingMode.full:
                    supplier_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                       source_child_id=supplier_key,
                                                                       mapping_type=MappingType.counterparty)
                    location_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                       source_child_id=location_key,
                                                                       mapping_type=MappingType.terminal
                                                                       )
                    product_source_id = mapper.get_gravitate_child_id(source_parent_id=configuration,
                                                                      source_child_id=product_key,
                                                                      mapping_type=MappingType.product)
                elif mapping_type == MappingMode.skip:
                    supplier_source_id = supplier_key
                    location_source_id = location_key
                    product_source_id = product_key
                else:
                    raise ValueError(f"Unsupported mapping type: {mapping_type}")
                if product_name is not None and location_name is not None and supplier_name is not None:
                    price_instrument_name = f"{product_name} @ {location_name} - {supplier_name}"
                else:
                    price_instrument_name = f"{product_source_id} @ {location_source_id} - {supplier_source_id}"
                price_instrument_source_string_id = f"{price_publisher_name} - {price_instrument_name}"
                dtos.append(PriceInstrumentDTO(
                    Name=price_instrument_name,
                    Abbreviation=price_instrument_name,
                    SourceIdString=price_instrument_source_string_id,
                    ProductLookup=PELookup(
                        SourceIdString=product_source_id,
                        SourceSystemId=int(source_system_id)
                    ),
                    LocationLookup=PELookup(
                        SourceIdString=location_source_id,
                        SourceSystemId=int(source_system_id)
                    ),
                    CounterPartyLookup=PELookup(
                        SourceIdString=supplier_source_id,
                        SourceSystemId=int(source_system_id)
                    )
                ))
            except (KeyError, ValueError) as e:
                mapping_failures.append(translated_row)
                logger.warning(f"Skipped record {translated_row} due to Key Error or Value Error: {e}")
                continue
            except Exception as uh:
                translation_failures.append(translated_row)
                logger.warning(f"Skipped record {translated_row} due to unhandled exception: {uh.args}")
                continue
            if dtos and price_publisher_name and source_system_id:
                yield PeBulkSyncIntegration(
                    IntegrationDtos=[
                        BulkSyncIntegrationDTO(
                            Name=price_publisher_name,
                            Abbreviation=price_publisher_name,
                            SourceIdString=price_publisher_name,
                            PriceInstrumentDTOs=dtos,
                            PriceTypeDTOs=[
                                PriceTypeDTO(
                                    PriceTypeMeaning=posting_type
                                )
                            ]
                        )
                    ],
                    SourceSystemId=int(source_system_id)
                )
            else:
                logger.warning("No valid records were processed to create an integration")


import os

from gcloud.aio.storage import Storage
from loguru import logger

from bb_integrations_lib.shared.model import FileReference
from bb_integrations_lib.protocols.pipelines import Step


class GCSExportFileStep(Step):
    def __init__(
            self,
            gcloud_storage: Storage,
            bucket: str,
            *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.storage = gcloud_storage
        self.bucket = bucket

    def describe(self) -> str:
        return "Exporting file to GCS bucket"

    async def execute(self, i: FileReference) -> FileReference:
        file_name = os.path.basename(i.file_path)
        with open(i.file_path, "rb") as f:
            file_data = f.read()

        if await self.storage.get_bucket(self.bucket).blob_exists(file_name):
            # If run twice and the file was already archived, we don't need to archive another copy.
            logger.debug(f"Blob '{file_name}' already exists in bucket '{self.bucket}', skipping upload")
            return i

        await self.storage.upload(self.bucket, file_name, file_data)
        return i



import asyncio
import json
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Any, AsyncIterator

import pytz
from loguru import logger

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.protocols.pipelines import GeneratorStep
from bb_integrations_lib.provider.ftp.client import FTPIntegrationClient
from bb_integrations_lib.secrets.credential_models import FTPCredential
from bb_integrations_lib.provider.ftp.model import FTPType
from bb_integrations_lib.shared.model import RawData, FileConfigRawData, ConfigMode, ConfigMatchMode
from bb_integrations_lib.util.utils import check_if_file_greater_than_date, file_exact_match


class SFTPFileConfigStep(GeneratorStep):

    def __init__(self, rita_client: GravitateRitaAPI,
                 ftp_client: FTPIntegrationClient | dict[str, FTPIntegrationClient],
                 mode: ConfigMode,
                 match_mode: ConfigMatchMode = ConfigMatchMode.Partial,
                 bucket_name: str | None = None,
                 config_name: str | None = None,
                 min_date: datetime = None,
                 min_date_tz: pytz.BaseTzInfo | None = None,
                 minutes_back: int | None = None,
                 strip_trailing_digits: bool = False,
                 use_server_modified_date: bool = False,
                 *args,
                 **kwargs) -> None:
        """
        Imports SFTP files based on the provided or discovered fileconfigs.

        :param rita_client: The RITA client to use to retrieve fileconfigs.
        :param ftp_client: The FTP client, or a dict of FTP clients with keys matching confignames, to use to retrieve
          data.
        :param mode: How the step should discover fileconfigs.
        :param match_mode: How the step should match fileconfigs to various properties of the files being scanned.
        :param bucket_name: The bucket name which holds fileconfigs, for FromBucket and ByName modes.
        :param config_name: The fileconfig name, if using ByName mode.
        :param min_date: Filter out files with a date before this.
        :param use_server_modified_date: If True and connection is SFTP, use server's file
          modification time instead of parsing date from filename. Falls back to filename
          parsing for FTP connections or if server date unavailable.
        """
        super().__init__(*args, **kwargs)
        self.rita_client = rita_client
        self.ftp_client = ftp_client
        self.mode = mode

        self.match_mode = match_mode
        self.strip_trailing_digits = strip_trailing_digits
        self.bucket_name = bucket_name
        self.config_name = config_name
        self.min_date = min_date
        self.min_date_tz = min_date_tz
        self.minutes_back = minutes_back
        self.use_server_modified_date = use_server_modified_date
        self.file_configs: dict[str, Any] = {}

        if self.mode == ConfigMode.FromBucket and not self.bucket_name:
            raise ValueError("Cannot use FromBucket mode without setting a bucket_name")
        if self.mode == ConfigMode.ByName and not self.bucket_name:
            raise ValueError("Cannot use ByName mode without setting a bucket_name")
        if self.mode == ConfigMode.ByName and not self.config_name:
            raise ValueError("Cannot use ByName mode without setting a config_name")

    async def load_file_configs(self):
        if self.mode == ConfigMode.AllFiltered:
            self.file_configs = await self.rita_client.get_file_configs()
        elif self.mode == ConfigMode.FromBucket:
            self.file_configs = await self.rita_client.get_fileconfigs_from_bucket(self.bucket_name)
        elif self.mode == ConfigMode.ByName:
            self.file_configs = await self.rita_client.get_fileconfig_by_name(self.bucket_name, self.config_name)
        logger.info(f"Loaded {len(self.file_configs)} fileconfigs: {self.config_name}")

    def describe(self) -> str:
        return "Importing SFTP files based on file configs"

    async def setup(self):
        if self.min_date is None:
            if self.pipeline_context and self.pipeline_context.max_sync:
                self.min_date = self.pipeline_context.max_sync.max_sync_date
            else:
                self.min_date = datetime.now(tz=self.min_date_tz)

        if self.minutes_back is not None:
            self.min_date -= timedelta(minutes=self.minutes_back)

    def file_modified_after_min_date(
        self,
        file_path: str,
        file_name: str,
        file_config: Any,
        ftp_client: FTPIntegrationClient
    ) -> bool:
        """
        Check if file was modified after min_date.

        Uses server modification time for SFTP when use_server_modified_date=True,
        otherwise falls back to parsing date from filename.

        :returns: True if file should be included (modified after min_date or no date check applies)
        """
        if not self.min_date:
            return True

        # Try server modification time (SFTP only)
        if self.use_server_modified_date:
            if ftp_client.credentials.ftp_type == FTPType.sftp:
                file_info = ftp_client.get_file_info(file_path)
                if file_info.last_modification_time:
                    file_mtime = datetime.fromtimestamp(file_info.last_modification_time, tz=pytz.UTC)
                    is_newer = file_mtime > self.min_date
                    if not is_newer:
                        logger.debug(f"Skipping {file_name}: server mtime {file_mtime} UTC < {self.min_date} {self.min_date_tz}")
                    return is_newer
            else:
                logger.warning(
                    f"use_server_modified_date=True but connection is {ftp_client.credentials.ftp_type}, "
                    "falling back to filename parsing"
                )

        # Fallback: parse date from filename
        if file_config.date_format != "":
            is_newer = check_if_file_greater_than_date(
                file_name, file_config.file_name, file_config.date_format,
                self.min_date, self.strip_trailing_digits, self.min_date_tz
            )
            if not is_newer:
                logger.debug(f"Skipping {file_name}: filename date <= {self.min_date}")
            return is_newer

        # No date check applies
        return True

    async def generator(self, i: Any) -> AsyncIterator[RawData]:
        await self.load_file_configs()

        for config_name, file_config in self.file_configs.items():
            if isinstance(self.ftp_client, dict):
                selected_ftp_client = self.ftp_client[config_name]
            else:
                selected_ftp_client = self.ftp_client
            logger.info(f"Scanning with fileconfig '{config_name}' in directory {file_config.inbound_directory}, looking for pattern '{file_config.file_name}' ({self.match_mode.name} match)")
            file_names = list(selected_ftp_client.list_files(file_config.inbound_directory))
            logger.info(f"Found {len(file_names)} file(s) in {file_config.inbound_directory}")
            matched_count = 0
            for idx, file_name in enumerate(file_names):
                if self.match_mode == ConfigMatchMode.Exact:
                    logger.info(f"Exact Matching file {file_name}")
                    if not file_exact_match(file_name, file_config.file_name):
                        logger.debug(f"Skipping file {file_name} due to not matching exactly to {file_config.file_name}")
                        continue
                elif self.match_mode == ConfigMatchMode.Partial:
                    if not file_config.file_name in file_name:
                        logger.debug(f"Skipping '{file_name}': does not contain '{file_config.file_name}'")
                        continue
                elif self.match_mode == ConfigMatchMode.ByExtension:
                    if not file_name.endswith(file_config.file_extension):
                        logger.debug(f"Skipping '{file_name}': does not end with '{file_config.file_extension}'")
                        continue
                file_path = str(Path(file_config.inbound_directory) / file_name)
                if not self.file_modified_after_min_date(file_path, file_name, file_config, selected_ftp_client):
                    continue
                matched_count += 1
                logger.info(f"Fetching {idx+1}/{len(file_names)}: {file_name}")
                rd = selected_ftp_client.download_file(file_path)
                self.pipeline_context.file_config = file_config
                yield FileConfigRawData(data=rd.data, file_name=rd.file_name, file_config=file_config)
            if matched_count == 0:
                logger.info(f"No files matching '{file_config.file_name}' found in {file_config.inbound_directory}")




if __name__ == "__main__":
    async def main():
        s = SFTPFileConfigStep(
            rita_client=GravitateRitaAPI(
                base_url="",
                username="",
                password=""
            ),
            ftp_client=FTPIntegrationClient(
                credentials=FTPCredential(
                    host="",
                    username="",
                    password="",
                    port=22,
                    ftp_type=FTPType.sftp
                ),
            ),
            mode=ConfigMode.ByName,
            config_name="my_config"
        )
        async for r in s.generator(None):
            print(r)

    asyncio.run(main())


from datetime import datetime, timedelta, UTC

import pandas as pd
from loguru import logger

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import GetOrderBolsAndDropsRequest


class TRTouchUpStep(Step):
    """Superclass for tank reading touchup steps"""
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

class NouriaDayDropsTRTouchUpStep(TRTouchUpStep):
    """
    Touch up a standardized tank readings report to add a "Order Number + Drop Index" column, which Nouria uses to get
    tank levels just before they dropped product.

    Note that the tenant/environment is hard coded to Nouria.

    """
    def __init__(
            self,
            sd_client: GravitateSDAPI,
            *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.sd_client = sd_client

    def describe(self) -> str:
        return "Touchup Nouria tank readings for day drops report"

    async def execute(self, i: pd.DataFrame) -> pd.DataFrame:
        df = i.copy()
        when = datetime.now(UTC)
        recent_drops_resp = await self.sd_client.get_bols_and_drops(GetOrderBolsAndDropsRequest(
            order_date_start=when - timedelta(days=1),
            order_date_end=when
        ))
        recent_drops = recent_drops_resp.json()
        dfg = df.groupby(["Store Number", "Tank Id"])
        out_series = []
        for bol in recent_drops:
            for drop_idx, drop in enumerate(bol["drops"]):
                site_no = str(drop["location"])
                tank_id = str(drop["tank_id"])
                try:
                    # Gets the readings rows for this drop's specific site/tank
                    group = dfg.get_group((site_no, tank_id))
                    # Keep reads that are before the before_stick timestamp, then get the index of the reading
                    # row that has the most recent (closest to before_stick_time) timestamp.
                    idx = group[group["Read Time"] < drop["before_stick_time"]]["Read Time"].idxmax()
                    # Copy it to avoid pandas grumbling
                    row = df.iloc[idx].copy()
                    row["Order Number + Drop Index"] = f"{bol['order_number']}-{drop_idx + 1}"
                    out_series.append(row)
                except KeyError:
                    logger.warning(f"{site_no}, tank {tank_id} in drops but not tank readings, skipping record")
                except ValueError as e:
                    logger.warning(f"{site_no}, tank {tank_id} could not be processed: {e}")
        # If there are no data rows, create a fake empty df with the same columns the actual data would have
        # (since there are no series items to infer headers from, the whole file would be empty)
        if len(out_series) == 0:
            empty_df = pd.DataFrame(data=None, columns=df.columns + ["Order Number + Drop Index"])
            return empty_df
        # Otherwise reconstitute the relevant rows (with newly added order number/drop index column)
        return pd.DataFrame.from_records(out_series).sort_values(["Store Number", "Tank Id"])



from typing import List, Dict, Any

from dateutil.parser import parse

from bb_integrations_lib.protocols.pipelines import Parser
from bb_integrations_lib.shared.model import MappingMode


class PriceSyncParser(Parser):
    def __init__(self, tenant_name: str, source_system: str | None = None):
        super().__init__(tenant_name=tenant_name, source_system=source_system)

    def __repr__(self) -> str:
        return "Parse prices from rows to Sync Price Rows in Pricing Engine."

    async def parse(self, data: List[Dict], mapping_type: MappingMode | None = None) -> Any:
        pass

    @staticmethod
    def is_rvp(rvp: str) -> bool:
        try:
            return float(rvp) > 0.0
        except (ValueError, TypeError):
            return False

    @staticmethod
    def format_rvp_product(product_key: str, rvp: str | None) -> str:
        if not rvp:
            return product_key
        rvp_str = str(rvp)
        if product_key.endswith(rvp_str):
            product_key = product_key[:-len(rvp_str)]
        product_key = product_key.rstrip('.')
        return f"{product_key}{float(rvp_str)}"

    @staticmethod
    def get_effective_to_date(effective_to_str: str | None,
                              effective_to_hrs_override: str | None,
                              effective_to_minutes_override: str | None) -> str | None:
        return PriceSyncParser.get_date_override(effective_to_str, effective_to_hrs_override,
                                                 effective_to_minutes_override)

    @staticmethod
    def get_effective_from_date(effective_from_str: str | None,
                                effective_from_hrs_override: str | None,
                                effective_from_minutes_override: str | None) -> str | None:
        return PriceSyncParser.get_date_override(effective_from_str, effective_from_hrs_override,
                                                 effective_from_minutes_override)

    @staticmethod
    def get_date_override(date_str: str | None,
                          date_to_hrs_override: str | None,
                          date_to_minutes_override: str | None) -> str | None:

        if not date_str:
            return date_str

        parsed_date = parse(date_str)

        if date_to_hrs_override and date_to_minutes_override:
            return parsed_date.replace(
                hour=int(date_to_hrs_override),
                minute=int(date_to_minutes_override),
                second=0
            ).isoformat()

        elif date_to_hrs_override:
            return parsed_date.replace(
                hour=int(date_to_hrs_override),
                minute=0,
                second=0
            ).isoformat()

        elif date_to_minutes_override:
            return parsed_date.replace(
                minute=int(date_to_minutes_override),
                second=0
            ).isoformat()

        else:
            return date_str



from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.mappers.prices.model import PricePublisher
from bb_integrations_lib.pipelines.parsers.price_engine.parse_accessorials_prices_parser import AccessorialPricesParser
from bb_integrations_lib.pipelines.steps.create_accessorials_step import BBDUploadAccessorialsStep
from bb_integrations_lib.pipelines.steps.exporting.pe_price_export_step import PEPriceExportStep
from bb_integrations_lib.protocols.pipelines import JobPipeline
from bb_integrations_lib.secrets import SecretProvider
from bb_integrations_lib.secrets.factory import APIFactory
from pydantic import BaseModel


class AccessorialPipelineConfig(BaseModel):
    price_publishers: list[PricePublisher]
    accessorial_date_timezone: str = "America/New_York"
    price_instrument_ids: list[int]
    source_system: str = "LCFS"
    hours_back: int = 24


class AccesorialsPriceTransformationPipeline(JobPipeline):
    @classmethod
    async def create(
            cls,
            config_id: str,
            rita_client: GravitateRitaAPI,
            pe_credentials: str,
            sd_credentials: str,
            secret_provider: SecretProvider,
            send_reports: bool = True
    ):
        api_factory = APIFactory(secret_provider)

        config = await rita_client.get_config_by_id(config_id)
        job_config: AccessorialPipelineConfig = AccessorialPipelineConfig.model_validate(config.config)

        steps = [
            {
                "id": "1",
                "parent_id": None,
                "step": PEPriceExportStep(
                    rita_client=rita_client,
                    pe_client=await api_factory.pe(pe_credentials),
                    price_publishers=job_config.price_publishers,
                    config_id=config_id,
                    hours_back=job_config.hours_back,
                    addl_endpoint_args={
                        "IsActiveFilterType": "ActiveOnly",
                        "PriceInstrumentIds": job_config.price_instrument_ids,
                        "IncludeSourceData": False,
                        "IncludeFormulaResultData": False
                    },
                    parser=AccessorialPricesParser,
                    parser_kwargs={
                        "source_system": job_config.source_system,
                        "timezone": job_config.accessorial_date_timezone,
                    }
                )
            },
            {
                "id": "2",
                "parent_id": "1",
                "step": BBDUploadAccessorialsStep(
                    sd_client=await api_factory.sd(sd_credentials),
                )
            }
        ]

        return cls(
            steps,
            rita_client=rita_client,
            pipeline_name=f"Custom Accessorials Price Integration",
            pipeline_config_id=config_id,
            secret_provider=secret_provider,
            send_reports=send_reports
        )



import re
from datetime import datetime, timedelta
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.provider.api.samsara.model import (
    GravitateELDRecord,
    HOSComplianceData,
    ShiftData,
)
from loguru import logger


class SamsaraTransformELDStep(Step):
    """Transform Samsara HOS data to Gravitate ELD format."""

    def __init__(
        self,
        *args,
        drive_max_hours: float = 11.0,
        on_duty_max_hours: float = 14.0,
        meters_to_miles: float = 0.000621371,
        ms_to_hours: float = 1 / (1000 * 60 * 60),
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.drive_max_hours = drive_max_hours
        self.on_duty_max_hours = on_duty_max_hours
        self.meters_to_miles = meters_to_miles
        self.ms_to_hours = ms_to_hours

    def describe(self) -> str:
        return "Transform Samsara data to Gravitate ELD format"

    async def execute(self, hos_data: list[dict]) -> list[dict]:
        """
        Transform Samsara HOS data into Gravitate ELD records.

        Args:
            hos_data: List of dicts from SamsaraRetrieveHOSDataStep

        Returns:
            List of Gravitate ELD record dicts ready for upload
        """
        eld_records = []

        for item in hos_data:
            try:
                record = self.transform_single(item)
                if record:
                    eld_records.append(record.model_dump(mode="json"))
            except Exception as e:
                samsara_id = item.get("samsara_id", "unknown")
                logger.warning(f"Failed to transform HOS data for driver {samsara_id}: {e}")
                continue

        logger.info(f"Transformed {len(eld_records)} ELD records")
        return eld_records

    def find_duty_period_start(
        self, clocks: dict, retrieve_timestamp: datetime
    ) -> datetime:
        """
        Calculate duty period start from Samsara shift clock.

        The shiftRemainingDurationMs represents time remaining in the 14-hour
        on-duty window since the shift began.

        Args:
            clocks: Samsara clocks dict with shift.shiftRemainingDurationMs
            retrieve_timestamp: Current retrieve timestamp

        Returns:
            The duty period start datetime (UTC)
        """
        shift_remaining_ms = clocks.get("shift", {}).get("shiftRemainingDurationMs", 0)
        shift_remaining_hours = shift_remaining_ms * self.ms_to_hours
        elapsed_shift_hours = self.on_duty_max_hours - shift_remaining_hours

        # Clamp to non-negative (in case of clock reset)
        elapsed_shift_hours = max(0, elapsed_shift_hours)

        return retrieve_timestamp - timedelta(hours=elapsed_shift_hours)

    def calculate_hours_from_clocks(
        self, clocks: dict
    ) -> tuple[float, float, float] | None:
        """
        Calculate drive, on-duty, and cycle remaining hours from Samsara clocks data.

        Args:
            clocks: Samsara clocks dict with drive, shift, and cycle keys

        Returns:
            Tuple of (total_drive_hours, total_on_duty_hours, cycle_remaining_hours)
            or None if clocks data is missing
        """
        if not clocks:
            return None

        drive_remaining_ms = clocks.get("drive", {}).get("driveRemainingDurationMs", 0)
        shift_remaining_ms = clocks.get("shift", {}).get("shiftRemainingDurationMs", 0)
        cycle_remaining_ms = clocks.get("cycle", {}).get("cycleRemainingDurationMs", 0)

        drive_remaining_hours = drive_remaining_ms * self.ms_to_hours
        shift_remaining_hours = shift_remaining_ms * self.ms_to_hours
        cycle_remaining_hours = cycle_remaining_ms * self.ms_to_hours

        total_drive_hours = self.drive_max_hours - drive_remaining_hours
        total_on_duty_hours = self.on_duty_max_hours - shift_remaining_hours

        return total_drive_hours, total_on_duty_hours, cycle_remaining_hours

    def parse_cycle_info_from_daily_log(
        self, daily_log: dict, cycle_remaining_hours: float
    ) -> tuple[float, float, str]:
        """
        Parse cycle information from daily log and calculate weekly on-duty hours.

        Args:
            daily_log: Samsara daily log dict
            cycle_remaining_hours: Remaining cycle hours from clocks

        Returns:
            Tuple of (weekly_max_hours, weekly_on_duty_hours, hos_cycle)
        """
        weekly_max_hours = 70.0  # Default to 70-hour
        hos_cycle = "8-day"  # Default

        driver_info = daily_log.get("driver", {})
        eld_settings = driver_info.get("eldSettings", {})
        rulesets = eld_settings.get("rulesets", [])

        if rulesets:
            cycle_str = rulesets[0].get("cycle", "")
            weekly_max_hours, hos_cycle = self.parse_cycle_string(cycle_str)

        weekly_on_duty_hours = weekly_max_hours - cycle_remaining_hours

        return weekly_max_hours, weekly_on_duty_hours, hos_cycle

    def transform_single(self, item: dict) -> GravitateELDRecord | None:
        """Transform a single driver's HOS data to Gravitate ELD format."""
        gravitate_driver = item["gravitate_driver"]
        samsara_id = item["samsara_id"]
        clocks = item["clocks"]
        daily_log = item["daily_log"]
        retrieve_timestamp = item["retrieve_timestamp"]

        hours_data = self.calculate_hours_from_clocks(clocks)
        if hours_data is None:
            logger.debug(f"Skipping driver {samsara_id}: no clocks data")
            return None

        total_drive_hours, total_on_duty_hours, cycle_remaining_hours = hours_data

        weekly_max_hours, weekly_on_duty_hours, hos_cycle = self.parse_cycle_info_from_daily_log(
            daily_log, cycle_remaining_hours
        )

        # Get distance traveled (convert meters to miles)
        distance_meters = daily_log.get("distanceTraveled", {}).get(
            "driveDistanceMeters", 0
        )
        total_miles = distance_meters * self.meters_to_miles

        # Calculate duty period start from shift clock
        duty_period_start = self.find_duty_period_start(clocks, retrieve_timestamp)
        duty_period_end = retrieve_timestamp  # Current time

        # Generate source_id with timestamp
        timestamp_str = retrieve_timestamp.strftime("%Y%m%dT%H%M%SZ")
        source_id = f"samsara_{samsara_id}_{timestamp_str}"

        # Get Gravitate driver ID
        gravitate_driver_id = gravitate_driver.get("_id") or gravitate_driver.get("id")

        return GravitateELDRecord(
            source_system="samsara",
            source_id=source_id,
            driver_id=gravitate_driver_id,
            duty_period_start=duty_period_start,
            duty_period_end=duty_period_end,
            shift_data=ShiftData(
                total_drive_hours=round(max(0, total_drive_hours), 2),
                total_on_duty_hours=round(max(0, total_on_duty_hours), 2),
                total_miles=round(total_miles, 2),
            ),
            hos_compliance_data=HOSComplianceData(
                drive_time_max_shift_hours=self.drive_max_hours,
                on_duty_max_shift_hours=self.on_duty_max_hours,
                weekly_on_duty_hours=round(max(0, weekly_on_duty_hours), 2),
                weekly_on_duty_max_hours=weekly_max_hours,
                hos_cycle=hos_cycle,
            ),
        )

    def parse_cycle_string(self, cycle_str: str) -> tuple[float, str]:
        """
        Parse Samsara cycle string like "USA 70 hour / 8 day" into (70.0, "8-day").

        Args:
            cycle_str: Samsara cycle string (e.g., "USA 70 hour / 8 day")

        Returns:
            Tuple of (max_hours, cycle_type) e.g., (70.0, "8-day")
        """
        # Default values
        max_hours = 70.0
        cycle_type = "8-day"

        # Try to extract hours (e.g., "70 hour" or "60 hour")
        hour_match = re.search(r"(\d+)\s*hour", cycle_str, re.IGNORECASE)
        if hour_match:
            max_hours = float(hour_match.group(1))

        # Try to extract days (e.g., "8 day" or "7 day")
        day_match = re.search(r"(\d+)\s*day", cycle_str, re.IGNORECASE)
        if day_match:
            cycle_type = f"{day_match.group(1)}-day"

        return max_hours, cycle_type



import asyncio
from datetime import datetime, UTC, timedelta
from typing import Dict, Any, Optional, Tuple, Iterable
from zoneinfo import ZoneInfo

import pandas as pd
from bson.raw_bson import RawBSONDocument
from loguru import logger
from pandas import DataFrame
from pymongo.asynchronous.database import AsyncDatabase
from pymongo.asynchronous.collection import AsyncCollection

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import ReadingQuery, ExportReadingsWindowMode


class BBDExportReadingsStep(Step):
    def __init__(self, rita_client: GravitateRitaAPI, sd_client: GravitateSDAPI, ims_database: AsyncDatabase,
                 reading_query: ReadingQuery, timezone: str = "UTC",
                 window_mode: ExportReadingsWindowMode = ExportReadingsWindowMode.HOURS_BACK, hours_back: float = 6,
                 batch_hours: float = 2, big_dataset: int = 750, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rita_client = rita_client
        self.sd_client = sd_client
        self.ims_database = ims_database

        self.reading_query = reading_query
        self.timezone = timezone
        self.window_mode = window_mode
        self.hours_back = hours_back
        self.batch_hours = batch_hours
        self.big_dataset = big_dataset

    def describe(self) -> str:
        return "Export tank readings from BBD with a customizable query"

    async def execute(self, _: Any) -> Tuple[Dict, Dict, Iterable]:
        # Compute start/end windows based on the provided configuration
        if self.window_mode == ExportReadingsWindowMode.HOURS_BACK or self.window_mode == ExportReadingsWindowMode.LATEST_ONLY:
            now = datetime.now(UTC)
            window_start = now - timedelta(hours=self.hours_back)
            window_end = now
        elif self.window_mode == ExportReadingsWindowMode.PREVIOUS_DAY:
            window_end = datetime.now(ZoneInfo(self.timezone)).replace(hour=0, minute=0, second=0, microsecond=0)
            window_start = window_end - timedelta(days=1)
        else:
            raise Exception(f"Unknown window mode {self.window_mode}")

        if self.window_mode == ExportReadingsWindowMode.HOURS_BACK:
            logger.info(
                f"Exporting tank readings from BBD for the last {self.hours_back} hours "
                f"(from {window_start} to {window_end})"
            )
        elif self.window_mode == ExportReadingsWindowMode.LATEST_ONLY:
            logger.info(
                f"Exporting ONLY the most recent tank reading from BBD for the last {self.hours_back} hours "
                f"(from {window_start} to {window_end}"
            )
        elif self.window_mode == ExportReadingsWindowMode.PREVIOUS_DAY:
            logger.info(f"Exporting tank readings from BBD for the previous day (from {window_start} to {window_end})")

        store_lkp, tank_lkp, filtered_stores = await self.get_sites_and_tank_ids_to_export()

        if len(filtered_stores) > self.big_dataset and self.reading_query.by_wildcard is None:
            raise ValueError(
                f"Large dataset ({len(filtered_stores)} stores) detected. "
                f"Must use wildcard query strategy (by_wildcard) for datasets > 500 stores."
            )

        use_or_filter = len(filtered_stores) <= self.big_dataset

        return await self.get_inventory(
            window_start=window_start,
            window_end=window_end,
            latest_only=self.window_mode == ExportReadingsWindowMode.LATEST_ONLY,
            store_lkp=store_lkp,
            tank_lkp=tank_lkp,
            filtered_stores=filtered_stores,
            use_or_filter=use_or_filter,
        )

    @staticmethod
    def collection_as_raw_bson(collection: AsyncCollection):
        return collection.with_options(
            codec_options=collection.codec_options.with_options(document_class=RawBSONDocument))

    async def get_sites_and_tank_ids_to_export(self, query: Optional[ReadingQuery] = None) -> Tuple[
        Dict, Dict, DataFrame]:
        query_to_use = query if query else self.reading_query
        stores = (await self.sd_client.get_all_stores(include_tanks=True)).json()
        filtered_stores = self.apply_reading_query_filter(stores=stores, query=query_to_use)
        filtered_store_numbers = filtered_stores["store_number"].to_list()
        store_lkp, tank_lkp = BBDExportReadingsStep.lkps(
            [store for store in stores if store["store_number"] in filtered_store_numbers])
        return store_lkp, tank_lkp, filtered_stores

    def apply_reading_query_filter(self, stores: list[Dict], query: Optional[ReadingQuery] = None) -> DataFrame:
        query_to_use = query if query else self.reading_query
        df = pd.DataFrame(stores)
        tanks_df = df.explode("tanks")
        tanks_df = tanks_df[~tanks_df["tanks"].isna()].reset_index()
        tanks_df["tank_id"] = tanks_df["tanks"].apply(lambda x: x["tank_id"])
        tanks_df["composite_key"] = tanks_df["store_number"] + ":" + tanks_df["tank_id"].astype(str)
        if query_to_use.by_wildcard is not None and query_to_use.by_wildcard == "*":
            return tanks_df[["store_number", "tank_id", "composite_key"]]
        mask = query_to_use.as_mask(tanks_df)
        filtered = tanks_df[mask]
        return filtered[["store_number", "tank_id", "composite_key"]]

    @staticmethod
    def ims_query_pairs(filtered_stores: DataFrame) -> list:
        pairs = []
        for _, row in filtered_stores.iterrows():
            pairs.append({
                "store_number": row["store_number"],
                "tank_id": str(row["tank_id"])
            })
        return pairs

    async def get_inventory(self,
                            window_start: datetime,
                            window_end: datetime,
                            latest_only: bool,
                            store_lkp: dict,
                            tank_lkp: dict,
                            filtered_stores: DataFrame | None = None,
                            use_or_filter: bool = True,  # Add this parameter
                            ) -> Tuple[Dict, Dict, Iterable]:
        """
        Get inventory / tank readings, filtering to the stores and tanks specified by reading_query.
        :param window_start: Include readings newer than or equal to this datetime.
        :param window_end: Include readings older than this datetime.
        :param latest_only: Whether only the latest reading within the window should be provided.
        :param store_lkp: Optional pre-fetched store lookup dict.
        :param tank_lkp: Optional pre-fetched tank lookup dict.
        :param filtered_stores: Optional pre-fetched filtered stores DataFrame.
        :param use_or_filter: Whether to include $or clause in aggregation for store/tank filtering.
        :return: A tuple containing a store lookup dict, tank lookup dict, and iterable of tank reading documents.
        """
        ims_pairs = BBDExportReadingsStep.ims_query_pairs(filtered_stores)
        collection = self.ims_database['tank_inventory_log']

        match_conditions: list[dict] = [
            {
                "read_time": {
                    "$gte": window_start,
                    "$lte": window_end
                }
            }
        ]

        if use_or_filter:
            match_conditions.append({
                                        "$or": ims_pairs
                                        })

        _query: list[dict] = [
            {
                "$match": {
                    "$and": match_conditions
                }
            },
            {
                "$sort": {
                    "read_time": 1
                }
            },
            {
                "$project": {
                    "tank_agent_name": 1,
                    "store_number": 1,
                    "run_time": 1,
                    "tank_id": 1,
                    "read_time": 1,
                    "product": 1,
                    "monitor_type": 1,
                    "volume": 1,
                }
            }
        ]
        # If we're getting only the most recent tank reading, aggregate by store and tank, then grab the latest
        # reading document from the lookback period.
        if latest_only:
            _query.append({
                "$group": {
                    "_id": {
                        "store_number": "$store_number",
                        "tank_id": "$tank_id"
                    },
                    "documents": {
                        "$last": "$$ROOT"
                    }
                }
            })

        cursor = await collection.aggregate(_query)
        results = await cursor.to_list()
        # Have to extract the latest document from the aggregate if we're grabbing the latest, as the query result
        # set is shaped differently.
        if latest_only:
            return store_lkp, tank_lkp, (r["documents"] for r in results)
        return store_lkp, tank_lkp, results

    @staticmethod
    def lkps(stores: Iterable) -> Tuple[Dict, Dict]:
        store_lkp = {}
        tank_lkp = {}
        for store in stores:
            store_number = store['store_number']
            tanks = store['tanks']
            for tank in tanks:
                tank_id = tank['tank_id']
                store_lkp[f"{store_number}"] = store
                tank_lkp[f"{store_number}:{tank_id}"] = tank
        return store_lkp, tank_lkp


if __name__ == "__main__":
    from pymongo import AsyncMongoClient

    async def main():
        export = BBDExportReadingsStep(
            rita_client=GravitateRitaAPI(
                base_url="",
                client_id="",
                client_secret=""
            ),
            sd_client=GravitateSDAPI(
                base_url="",
                client_id="",
                client_secret=""
            ),
            ims_database=AsyncMongoClient("mongodb conn str")["db_name"],
            reading_query=ReadingQuery(
                by_store_numbers=["100101"]
            )
        )
        return await export.execute("Majors")

    readings = asyncio.run(main())



import json
from time import sleep
from typing import Dict, List

from loguru import logger

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.models.pipeline_structs import BBDUploadResult
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.util.utils import CustomJSONEncoder


class BBDUploadAccessorialsStep(Step):
    def __init__(
            self,
            sd_client: GravitateSDAPI,
            buffer: float = 0.5,
            chuk_size: int = 1000,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.sd_client = sd_client
        self.buffer = buffer
        self.chuk_size = chuk_size

    def describe(self) -> str:
        return "Upload Accessorials to BBD"

    async def execute(self, accessorials: List[Dict]) -> BBDUploadResult:
        logs = {"requests": [], "responses": [], "errors": []}
        try:
            total_accessorials = len(accessorials)
            succeeded = []
            failed_items = []

            for idx, accessorial in enumerate(accessorials):
                logs["requests"].append(accessorial)
                try:
                    resp = await self.sd_client.call_ep("freight/accessorial/automatic/rate/create", json=accessorial)
                    resp.raise_for_status()
                    _json = resp.json()
                    sleep(self.buffer)
                    succeeded.append(accessorial)
                    logger.info(f"Accessorials uploaded successfully: {idx + 1} of {total_accessorials}")
                    logs["responses"].append({"response": _json, "request": accessorial})
                except Exception as e:
                    logs["errors"].append({"record": accessorial, "error": f"Error uploading accessorials: {str(e)} {e.response.content}"})
                    failed_items.append(accessorial)
                    continue

            self.pipeline_context.included_files["accessorials data upload"] = json.dumps(logs, cls=CustomJSONEncoder)
            return BBDUploadResult(
                succeeded=len(succeeded),
                failed=len(failed_items),
                succeeded_items=succeeded
            )

        except Exception as e:
            logger.exception(f"Unable to upload | {e}")
            raise e



from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import RawData


class SaveRawDataToDiskStep(Step):
    """Save a RawData object to disk in the current working directory."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def describe(self) -> str:
        return "Save a RawData object to the current working directory"

    async def execute(self, i: RawData | list[RawData]) -> None:
        items = i if isinstance(i, list) else [i]
        for item in items:
            with open(item.file_name, "wb") as f:
                f.write(item.data.read() if hasattr(item.data, "read") else item.data)



from datetime import datetime
from enum import StrEnum
from typing import List

from pydantic import BaseModel


class ERPStatus(StrEnum):
    """Copied from supply_and_dispatch_aio 2025-06-17"""
    sent = "sent"
    unsent = "unsent"
    errors = "errors"

class BackofficeERP(BaseModel):
    """Copied from supply_and_dispatch_aio 2025-06-17"""
    status: ERPStatus = ERPStatus.unsent
    time_sent: datetime | None = None
    errors: List = []



from bb_integrations_lib.provider.api.cargas.model import CreateWholesaleTicketRequest, CreateWholesaleLineRequest, \
    CreateWholesaleFeeLineRequest


class CargasClient():
    def __init__(self,
                 api_key: str,
                 base_url: str = " https://cargase.aeroenergy.com/cargasenergy/API"):
        self.base_url = base_url
        self.api_key = api_key

    @property
    def custom_headers(self) -> dict:
        return {
            'APIKey': self.api_key,
            'Content-Type': 'application/json'
        }

    async def create_wholesale_ticket(self, data: CreateWholesaleTicketRequest): ...

    async def create_wholesale_line(self, data: CreateWholesaleLineRequest): ...

    async def create_wholesale_fee_line(self, data: CreateWholesaleFeeLineRequest): ...



from dataclasses import dataclass
from datetime import datetime
from itertools import groupby
from zoneinfo import ZoneInfo

from bb_integrations_lib.models.sd.bols_and_drops import Freight
from bb_integrations_lib.util.utils import lookup


@dataclass
class Entity:
    use_source_id: bool
    source_id_key: str
    fallback_field: str = "name"


def safe_get_extra_data(obj: dict, key: str):
    """Get a value from an object's extra_data dict."""
    if "extra_data" in obj:
        return obj.get("extra_data", {}).get(key)
    return None


def to_timezone(dt: datetime, zone: str) -> datetime | None:
    """Convert a datetime to the given timezone from UTC."""
    if not dt:
        return None
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=ZoneInfo("UTC"))
    try:
        return dt.astimezone(ZoneInfo(zone)).replace(tzinfo=None)
    except OverflowError:
        return dt.replace(tzinfo=None)


def get_entity_id(entities: dict[str, Entity], entity_type: str, entity_data: dict) -> str | None:
    """Get the identifier for an entity based on the entity config."""
    entity_config = entities.get(entity_type)
    if not entity_config:
        return None
    if entity_config.use_source_id:
        return safe_get_extra_data(entity_data, entity_config.source_id_key)
    return entity_data.get(entity_config.fallback_field)


def aggregate_accessorials(freight: Freight | None, agg_key: str = "total", agg_group: str = "subtype") -> dict:
    """Aggregate accessorial FreightTransactions by a grouping key, summing the agg_key values."""
    if not freight:
        return {}
    accessorials = [t for t in freight.transactions if t.type == "Accessorial"]
    sorted_details = sorted(accessorials, key=lambda x: getattr(x, agg_group, None) or "")
    aggregates = {}
    for key, group in groupby(sorted_details, key=lambda x: getattr(x, agg_group, None)):
        group_list = list(group)
        total = sum(getattr(detail, agg_key, None) or 0 for detail in group_list)
        label = (key or "").strip()
        aggregates[label] = {"total": total}
    return aggregates


def build_accessorials_strings(freight: Freight | None, delimiter: str = ",") -> tuple[str, str]:
    """
    Aggregate accessorial transactions from a Freight model
    and return joined subtypes and amounts strings.
    """
    aggregated = aggregate_accessorials(freight)
    subtypes = list(aggregated.keys())
    amounts = [str(v.get("total")) for v in aggregated.values()]
    subtypes_str = delimiter.join(subtypes) if subtypes else ""
    amounts_str = delimiter.join(amounts) if amounts else ""
    return subtypes_str, amounts_str


def build_lookups(
    counterparties: list[dict],
    locations: list[dict],
    stores: list[dict],
    products: list[dict],
    drivers: list[dict],
) -> dict:
    """Build the standard set of entity lookups used by order parsers."""
    cp_lkp = lookup(counterparties, lambda x: x.get("id"))
    return {
        "cp_lkp": cp_lkp,
        "loc_lkp": lookup(locations, lambda x: x.get("id")),
        "store_lkp": lookup(stores, lambda x: x.get("_id")),
        "prod_lkp": lookup(products, lambda x: x.get("id")),
        "driver_lkp": lookup(drivers, lambda x: x.get("id")),
        "store_lkp_by_number": lookup(stores, lambda x: x.get("store_number")),
        "internal_cp_names": [
            c.get("name") for _id, c in cp_lkp.items() if "Internal" in (c.get("types") or [])
        ],
    }



from bb_integrations_lib.protocols.pipelines import Step, Input
from loguru import logger


class EchoStep(Step[Input, Input]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def describe(self):
        return "Echo step input at debug priority"

    async def execute(self, i: Input) -> Input:
        logger.debug(i)
        return i



import json
from typing import AsyncGenerator, Optional, cast
from typing import override

import dateutil
from babel.numbers import parse_decimal, NumberFormatError
from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from dateutil.tz import gettz
from loguru import logger

from bb_integrations_lib.mappers.rita_mapper import RitaMapper, AsyncMappingProvider
from bb_integrations_lib.models.rita.issue import IssueCategory
from bb_integrations_lib.models.rita.mapping import MappingType
from bb_integrations_lib.protocols.flat_file import TankReading, TankMonitorType
from bb_integrations_lib.protocols.pipelines import Parser
from bb_integrations_lib.shared.model import MappingMode
from bb_integrations_lib.util.utils import lookup

tzmapping = {
    'EST': gettz("US/Eastern"),
    'EDT': gettz("US/Eastern"),
    'CST': gettz("US/Central"),
    'CDT': gettz("US/Central"),
    'MST': gettz("US/Mountain"),
    'MDT': gettz("US/Mountain"),
    'PST': gettz("US/Pacific"),
    'PDT': gettz("US/Pacific"),
}


class TankReadingParser(Parser):
    def __init__(
            self,
            sd_client: GravitateSDAPI,
            source_system: str | None = None,
            mapping_provider: Optional[AsyncMappingProvider] = None,
            included_payload: Optional[dict] = None,
            verbose: bool = False,
            *args,
            **kwargs
    ):
        super().__init__(source_system, mapping_provider)
        self.sd_client = sd_client
        self.verbose = verbose
        self.included_payload = included_payload or {}
        self.mapper: Optional[RitaMapper] = None

    @override
    async def parse(self, data: list[dict], mapping_type: MappingMode | None = None) -> AsyncGenerator[
        TankReading, None]:
        if mapping_type is None:
            logger.warning("TankReadingParser.parse mapping_type is None, defaulting to skip")
            mapping_type = MappingMode.skip
        sd_store_lkp = await self.get_store_lkp()
        self.mapper = await self.load_mapper()
        preparsed_records = self.preparse(data, mapping_type)

        for rec in preparsed_records:
            with logger.catch(message=f"Skipped record {rec} due to error"):
                store_id = rec.get("site_id")
                sd_store = sd_store_lkp.get(store_id , {})
                store_tz = sd_store.get("timezone")
                yield TankReading(
                    store=store_id,
                    date=rec.get("reading_time"),
                    monitor_type=TankMonitorType.bbd,
                    timezone=rec.get("timezone") or store_tz,
                    volume=rec.get("volume"),
                    tank=rec.get("tank_id"),
                    payload=self.included_payload
                )
    async def get_store_lkp(self):
        stores = await self.sd_client.all_stores()
        return lookup(stores.json(), lambda x: x.get("store_number"))

    def preparse(self, records: list[dict], mapping_type: MappingMode) -> list:
        """Perform basic sanity checking on records and map tank and site ids, if applicable."""
        parsed_records = []
        mapping_failures = []
        for translated in records:
            try:
                translated_volume = translated.get("volume")
                if translated_volume == 'nan':
                    if self.verbose:
                        logger.warning(f"Skipped record {translated} due to NaN volume.")
                    continue
                try:
                    if isinstance(translated_volume, str):
                        trans_vol_decimal = float(parse_decimal(translated_volume, locale="en_US"))
                    else:
                        trans_vol_decimal = float(translated_volume)

                    translated["volume"] = trans_vol_decimal
                except NumberFormatError:
                    if self.verbose:
                        logger.warning(
                            f"Skipped record {translated} due to invalid volume value '{translated_volume}'.")
                    continue
                except TypeError:
                    if self.verbose:
                        logger.warning(
                            f"Skipped record {translated} due to invalid volume value '{translated_volume}'."
                        )
                    continue
                if trans_vol_decimal < 0:
                    if self.verbose:
                        logger.warning(f"Skipped record {translated} due to negative volume.")
                    continue
                if translated.get("tank_id") == "nan":
                    if self.verbose:
                        logger.warning(f"Skipped record {translated} due to NaN tank.")
                    continue
                if not translated.get("reading_time"):
                    logger.warning(f"Skipped record {translated} due to missing date")
                    continue
                try:
                    date_parsed = dateutil.parser.parse(translated.get("reading_time"), tzinfos=tzmapping)
                    translated["reading_time"] = date_parsed.isoformat()
                except Exception as parse_error:
                    logger.warning(f"Skipped record {translated} due to date parsing error: {parse_error}")
                    continue

                if mapping_type == MappingMode.skip:
                    parsed_records.append(translated)
                elif mapping_type == MappingMode.partial or mapping_type == MappingMode.full:
                    try:
                        site_id = translated["site_id"]
                        tank_id = translated["tank_id"]
                        mapped_site_ids = self.mapper.get_gravitate_parent_ids(site_id, MappingType.site)
                        mapped_tank_ids = self.mapper.get_gravitate_child_ids(site_id, tank_id.strip(), MappingType.tank)
                        for site_id in mapped_site_ids:
                            for tank_id in mapped_tank_ids:
                                translated["site_id"] = site_id
                                translated["tank_id"] = tank_id
                                parsed_records.append(translated)
                    except (KeyError, ValueError) as e:
                        if mapping_type == MappingMode.partial:
                            parsed_records.append(translated)
                        else:
                            raise e

            except (KeyError, ValueError) as e:
                logger.warning(f"Skipped record {translated} due to error: {e}")
                mapping_failures.append(translated)
        if len(mapping_failures) > 0:
            self.record_issue(
                key_suffix="mapping_errors",
                name=f"Mapping errors",
                category=IssueCategory.TANK_READING,
                problem_short=f"{len(mapping_failures)} rows failed to map",
                problem_long=json.dumps(mapping_failures)
            )
        return parsed_records



class MappingNotFoundException(Exception):
    pass

class StepInitializationError(Exception):
    pass

class FileParsingError(Exception):
    """Raised when file parsing fails."""
    pass

class MapperLoadError(Exception):
    """Raised when loading the RITA mapper fails."""
    pass

class ConfigNotFoundError(Exception):
    """Raised when a configuration is not found."""
    pass

class ConfigValidationError(Exception):
    """Raised when configuration validation fails."""
    pass

class StepConfigValidationError(Exception):
    """Raised when a step configuration validation fails."""
    pass


import httpx
from httpx import Request, Response

from bb_integrations_lib.provider.api.keyvu.model import KeyVuDeliveryPlan, default_serialization_options


class KeyVuAuth(httpx.Auth):
    def __init__(self, token):
        self.token = token

    def auth_flow(self, request: Request):
        request.headers["Keyvu-Api-Key"] = self.token
        yield request


class KeyVuClient:
    def __init__(self, api_key: str):
        self.base_url = ""
        self.api_key = api_key
        self.client = httpx.AsyncClient(auth=KeyVuAuth(self.api_key))

    def build_url(self, tail: str) -> str:
        return f"{self.base_url}/{tail}"

    async def upload_deliveryplan(self, data: KeyVuDeliveryPlan, override_filename: str | None = None) -> Response:
        return await self.client.post(
            url=self.build_url("upload/deliveryplan"),
            content=data.to_xml(**default_serialization_options),
            headers={"filename": override_filename} if override_filename else {}
        )



import math
from datetime import datetime, timedelta

from bb_integrations_lib.shared.model import ConfigMatchMode, ConfigMode, MappingMode
from pydantic import BaseModel, field_validator, Field
from dateutil.parser import parse
from bb_integrations_lib.shared.shared_enums import PriceType


class SupplyOwnerConfig(BaseModel):
    gravitate_id: str
    extend_by_days: int = 3
    expire_in_hours: int = 24

class DTNIntegrationConfig(BaseModel):
    config_name: str
    ftp_credentials: str
    file_match_mode: ConfigMatchMode = ConfigMatchMode.Partial
    config_match_mode: ConfigMode = ConfigMode.ByName
    mapping_mode: MappingMode = MappingMode.full
    supply_owners: list[SupplyOwnerConfig]

class DTNPriceRecord(BaseModel):
    supplier: str
    terminal: str
    product: str
    price: float
    source_system: str = "DTN"
    effective_from_date: datetime
    brand: str | None = None
    supply_owner: str | None = Field(default=None, alias="supply owner")
    price_type: PriceType | None = Field(default=PriceType.rack, alias="price type")
    contract: str | None = None
    model_config = {'extra': 'ignore', 'populate_by_name': True}

    @field_validator('supplier', 'terminal', 'product', 'price', mode='before')
    @classmethod
    def reject_nan(cls, v, info):
        if v is None:
            raise ValueError(f'{info.field_name} cannot be None')
        if isinstance(v, str) and v.lower() == 'nan':
            raise ValueError(f'{info.field_name} cannot be NaN')
        if isinstance(v, float) and math.isnan(v):
            raise ValueError(f'{info.field_name} cannot be NaN')
        return v

    @field_validator('effective_from_date', mode='before')
    @classmethod
    def parse_date(cls, v, info):
        if isinstance(v, str):
            try:
                return parse(v)
            except (ValueError, TypeError):
                raise ValueError(f'{info.field_name} must be a valid datetime string')
        return v

    @property
    def source_id(self) -> str:
        parts = [self.supplier, self.terminal, self.product, self.source_system, self.supply_owner]
        if self.brand:
            parts.append(self.brand)
        return "|".join(parts)

    @property
    def map_key(self):
        return {
            "supplier": self.supplier,
            "terminal": self.terminal,
            "product": self.product,
            "brand": self.brand,
        }

    def add_days(self, days: int) -> datetime:
        """Return effective_from_date plus the specified days."""
        return self.effective_from_date + timedelta(days=days)

    def add_hours(self, hours: int) -> datetime:
        """Return effective_from_date plus the specified hours."""
        return self.effective_from_date + timedelta(hours=hours)


from asyncio import Future, wait_for
from uuid import uuid4

from faststream.kafka.fastapi import KafkaRouter, Context

from bb_integrations_lib.models.rita.workers import WorkerRequest, WorkerResponse
from bb_integrations_lib.workers.topics import WORKER_IMMEDIATE_REQUEST, WORKER_IMMEDIATE_RESPONSE, build_topic


class RPCWorker:
    """An async interface for sending awaitable remote procedure calls over Kafka to workers."""
    def __init__(self, router: KafkaRouter, namespace: str) -> None:
        self.responses: dict[str, Future[WorkerResponse]] = {}
        self.reply_topic = build_topic(namespace, WORKER_IMMEDIATE_RESPONSE)
        self.send_topic = build_topic(namespace, WORKER_IMMEDIATE_REQUEST)

        self.router = router
        self.subscriber = self.router.subscriber(self.reply_topic)
        self.subscriber(self._handle_responses)

    async def _handle_responses(self, response: WorkerResponse, message = Context("message")) -> None:
        if future := self.responses.pop(message.correlation_id, None):
            future.set_result(response)

    async def request(
            self,
            runnable_name: str,
            tenant_name: str,
            send_topic: str | None = None,
            reply_topic: str | None = None,
            runnable_kwargs: dict | None = None,
            timeout: float = 30.0,
    ) -> WorkerResponse:
        """Send a request to be executed on a worker, but wait for the response with a configurable timeout."""
        send_topic = send_topic or self.send_topic
        reply_topic = reply_topic or self.reply_topic

        correlation_id = str(uuid4())
        future = self.responses[correlation_id] = Future[WorkerResponse]()

        # Use an originator of "other" so that status reports don't get reported to the backend task tracker.
        request = WorkerRequest(runnable_name=runnable_name, tenant_name=tenant_name, originator="other", runnable_kwargs=runnable_kwargs)
        await self.router.broker.publish(request, topic=send_topic, reply_to=reply_topic, correlation_id=correlation_id)

        try:
            response: WorkerResponse= await wait_for(future, timeout=timeout)
        except TimeoutError:
            _ = self.responses.pop(correlation_id, None)
            raise
        return response



from abc import abstractmethod, ABC
from collections import defaultdict
from enum import Enum
from typing import Optional, Union

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.rita.mapping import Map, MappingType, Children, CompositeMapKey
from bb_integrations_lib.secrets import RITACredential
from bb_integrations_lib.util.cache.custom_ttl_cache import CustomAsyncTTLCache


class MappingsNotLoadedException(Exception):
    pass


class RitaMapperKeyError(KeyError):
    def __init__(self, *args, from_key: str | CompositeMapKey, **kwargs):
        super().__init__(*args, **kwargs)

        self.from_key = from_key


class FastMapDirection(Enum):
    source_to_grav = 1
    grav_to_source = 2


class FastMapKey:
    """
    A hashable key that can be used to identify mappings quickly if enough properties are known.
    Supports both simple string IDs and composite keys for entity identity (from_id).

    Note: parent_id is always a simple string. Composite keys are for entity identity resolution,
    not for hierarchical relationships. Children mappings only support string identifiers.
    """
    from_id: Union[str, CompositeMapKey]
    type: Optional[MappingType] = None
    parent_id: Optional[str] = None
    direction: FastMapDirection

    def __init__(
        self,
        from_id: Union[str, CompositeMapKey],
        type: Optional[MappingType],
        parent_id: Optional[str],
        direction: FastMapDirection
    ):
        self.from_id = from_id
        self.type = type
        # parent_id is always a simple string (hierarchy uses string IDs)
        self.parent_id = parent_id
        self.direction = direction

        # Cache normalized values for hash/equality
        self._from_id_normalized = self._normalize_from_id(self.from_id)
        self._parent_id_normalized = self.parent_id or ""

    @staticmethod
    def _normalize_from_id(value: Union[str, CompositeMapKey]) -> str:
        """Convert from_id to hashable string representation."""
        if isinstance(value, CompositeMapKey):
            return value.to_cache_key()
        return value

    def __hash__(self):
        return hash((
            self._from_id_normalized,
            self.type,
            self._parent_id_normalized,
            self.direction.value
        ))

    def __eq__(self, other):
        if not isinstance(other, FastMapKey):
            return False
        return (
            self._from_id_normalized == other._from_id_normalized and
            self.type == other.type and
            self._parent_id_normalized == other._parent_id_normalized and
            self.direction == other.direction
        )

    def __repr__(self):
        type_name = self.type.name if self.type else None
        parent_id_str = self._parent_id_normalized or None
        return f"FastMap<from_id={self._from_id_normalized}, type_name={type_name}, parent_id={parent_id_str}, direction={self.direction.name}>"


class MappingProvider(ABC):
    """An abstract base class for implementing synchronous mapping providers for the Mapper class."""

    def __init__(self):
        pass

    @abstractmethod
    def get_mappings_by_source_system(self, source_system: str) -> list[Map]:
        pass


class AsyncMappingProvider(ABC):
    """An abstract base class for implementing asynchronous mapping providers for the Mapper class."""

    def __init__(self):
        pass

    @abstractmethod
    async def get_mappings_by_source_system(self, source_system: str) -> list[Map]:
        pass


class MockMappingProvider(MappingProvider):
    """A mapping provider that uses a constant list of mappings, intended for testing."""

    def __init__(self, mappings: list[Map]):
        super().__init__()
        self.mappings = mappings

    def get_mappings_by_source_system(self, source_system: str) -> list[Map]:
        """Returns mappings originally passed to init, filtered by source system."""
        return list(filter(lambda m: m.source_system == source_system, self.mappings))


class MockAsyncMappingProvider(AsyncMappingProvider):
    """A mapping provider that uses a constant list of mappings, intended for testing."""

    def __init__(self, mappings: list[Map]):
        super().__init__()
        self.mappings = mappings

    async def get_mappings_by_source_system(self, source_system: str) -> list[Map]:
        """Returns mappings originally passed to init, filtered by source system."""
        return list(filter(lambda m: m.source_system == source_system, self.mappings))


class RitaAPIMappingProvider(AsyncMappingProvider):
    """Accesses the RITA API to retrieve mappings, providing them to a Mapper."""

    def __init__(self, rita_client: GravitateRitaAPI):
        super().__init__()
        self.rita_client = rita_client

    async def from_credential(self, credential: RITACredential):
        self.rita_client = GravitateRitaAPI.from_credential(credential)

    async def get_mappings_by_source_system(self, source_system: str) -> list[Map]:
        return await self.rita_client.get_mappings_by_source_system(source_system)


class RitaAPICachedMappingProvider(RitaAPIMappingProvider):
    def __init__(
            self,
            rita_client: GravitateRitaAPI,
            ttl_seconds: int = 120
    ):
        super().__init__(rita_client)
        self.ttl_seconds = ttl_seconds
        self.cache: CustomAsyncTTLCache = CustomAsyncTTLCache(verbose=False)

    async def get_mappings_by_source_system(self, source_system: str) -> list[Map]:
        return await self.cache.get_or_set(
            source_system, self.ttl_seconds, super().get_mappings_by_source_system, source_system
        )


class RitaMapperCore(ABC):
    """
    Provides the core logic of mapping lookups. Entirely synchronous; the concrete implementations can use these
    methods directly, and should generally only implement loading mappings.
    """

    def __init__(self):
        self.loaded = False
        self.fast_maps = defaultdict(list)

    def _raise_if_not_loaded(self):
        if not self.loaded:
            raise MappingsNotLoadedException("Mappings used before getting loaded")

    def _load_mappings(self, mappings: list[Map]):
        """
        Load the mappings from mappings into the mapper's internal data structures, making the get_ calls available
        for use. Supports both simple string IDs and composite keys for entity identity.

        Note: parent_id in FastMapKey is always a string. If a parent Map has a composite source_id,
        it's converted to string for hierarchical lookups.
        """

        def _get_identifier(map_obj: Map | Children, direction: FastMapDirection):
            """Extract the appropriate identifier based on a map direction."""
            if direction == FastMapDirection.grav_to_source:
                return map_obj.gravitate_id
            return map_obj.source_id

        def _get_parent_id_str(parent: Map | None, direction: FastMapDirection) -> str | None:
            """Get parent identifier as string (hierarchy always uses string IDs)."""
            if parent is None:
                return None
            parent_id = _get_identifier(parent, direction)
            # Convert composite keys to string for parent_id (hierarchy uses strings)
            if isinstance(parent_id, CompositeMapKey):
                return parent_id.to_cache_key()
            return parent_id

        def _reg_map(map: Map | Children, direction: FastMapDirection, parent: Map | None = None):
            from_id = _get_identifier(map, direction)
            parent_id = _get_parent_id_str(parent, direction)
            map_key = FastMapKey(from_id, map.type, parent_id, direction)
            self.fast_maps[map_key].append(map)

        for direction in FastMapDirection:
            for m in mappings:
                for child in m.children or []:
                    _reg_map(child, direction, m)
                _reg_map(m, direction, None)

        self.loaded = True

    # noinspection PyUnreachableCode
    def get_mappings(self, from_id: str, direction: FastMapDirection, mapping_type: MappingType,
                     parent_from_id: str | None = None) -> list[Map | Children]:
        """
        Retrieve mappings meeting the criteria. Base method for many of the get_gravitate/source* methods - prefer using
        those for a more expressive syntax.

        :arg str from_id: The original value we're looking to map. If direction is source_to_grav, this matches on the
          source_id field; if direction is grav_to_source, this matches on the gravitate_id field.
        :arg FastMapDirection direction: The direction of the mapping (gravitate->source or source->gravitate).
        :arg MappingType mapping_type: The type of the mapping.
        :arg str | None parent_from_id: The from_id of the child mapping's parent, if searching for a child. To get
          parent mappings instead, use None (the default).
        :raises MappingsNotLoadedException: If mappings have not been loaded yet.
        """
        self._raise_if_not_loaded()
        if mapping_type is not None and not isinstance(mapping_type, MappingType):
            raise TypeError(
                "mapping_type must be of type bb_integrations_lib.models.rita.mapping.MappingType "
                "(check whether you imported the right MappingType)"
            )

        return self.fast_maps.get(FastMapKey(from_id, mapping_type, parent_from_id, direction), [])

    def get_mappings_guaranteed(
            self, from_id: str, direction: FastMapDirection, mapping_type: MappingType,
            parent_from_id: str | None = None
    ) -> list[Map | Children]:
        """Like get_mappings, but additionally raises a RitaMapperKeyError if no mappings are found."""
        self._raise_if_not_loaded()
        mappings = self.get_mappings(from_id, direction, mapping_type, parent_from_id)
        if not mappings:
            from_name = "gravitate" if direction == FastMapDirection.grav_to_source else "source"
            child_slug = f", parent_{from_name}_id='{parent_from_id}'" if parent_from_id else ""
            mapping_type_name = "None" if mapping_type is None else mapping_type.name
            raise RitaMapperKeyError(
                f"No mappings found for {from_name}_id='{from_id}'{child_slug} with type='{mapping_type_name}'",
                from_key=from_id,
            )
        return mappings

    def get_single_mapping_guaranteed(
            self, from_id: str, direction: FastMapDirection, mapping_type: MappingType,
            parent_from_id: str | None = None
    ) -> Map | Children:
        """
        Like get_mappings_guaranteed, but additionally raises a RitaMapperKeyError if more than 1 mapping is found.
        """
        mappings = self.get_mappings_guaranteed(from_id, direction, mapping_type, parent_from_id)
        if len(mappings) > 1:
            from_name = "gravitate" if direction == FastMapDirection.grav_to_source else "source"
            child_slug = f", parent_{from_name}_id='{parent_from_id}'" if parent_from_id else ""
            raise RitaMapperKeyError(
                f"Too many mappings ({len(mappings)}) for {from_name}_id='{from_id}'{child_slug} "
                f"with type='{mapping_type.name}', expected 1",
                from_key=from_id,
            )
        return mappings[0]

    def get_gravitate_parent_id(
        self,
        source_id: str | CompositeMapKey,
        mapping_type: MappingType | None
    ) -> str | CompositeMapKey:
        return self.get_single_mapping_guaranteed(source_id, FastMapDirection.source_to_grav, mapping_type).gravitate_id

    def get_gravitate_child_id(
        self,
        source_parent_id: str,
        source_child_id: str,
        mapping_type: MappingType | None
    ) -> str:
        """
        Get gravitate_id for a child mapping.

        Note: Child lookups use string IDs only. Children mappings don't support composite keys.
        For composite key lookups (without hierarchy), use get_gravitate_id_by_composite().
        """
        return self.get_single_mapping_guaranteed(
            source_child_id, FastMapDirection.source_to_grav, mapping_type, source_parent_id
        ).gravitate_id

    def get_source_parent_id(
        self,
        gravitate_id: str | CompositeMapKey,
        mapping_type: MappingType | None
    ) -> str | CompositeMapKey:
        return self.get_single_mapping_guaranteed(gravitate_id, FastMapDirection.grav_to_source, mapping_type).source_id

    def get_source_child_id(
        self,
        gravitate_parent_id: str,
        gravitate_child_id: str,
        mapping_type: MappingType | None
    ) -> str:
        """
        Get source_id for a child mapping.

        Note: Child lookups use string IDs only. Children mappings don't support composite keys.
        For composite key lookups (without hierarchy), use get_source_id_by_composite().
        """
        return self.get_single_mapping_guaranteed(
            gravitate_child_id, FastMapDirection.grav_to_source, mapping_type, gravitate_parent_id
        ).source_id

    def get_gravitate_parent_ids(
        self,
        source_id: str | CompositeMapKey,
        mapping_type: MappingType | None
    ) -> list[str | CompositeMapKey]:
        return [
            x.gravitate_id for x in
            self.get_mappings_guaranteed(source_id, FastMapDirection.source_to_grav, mapping_type)
        ]

    def get_gravitate_child_ids(
        self,
        source_parent_id: str,
        source_child_id: str,
        mapping_type: MappingType | None
    ) -> list[str]:
        """
        Get all gravitate_ids for child mappings matching the criteria.

        Note: Child lookups use string IDs only. Children mappings don't support composite keys.
        """
        return [
            x.gravitate_id for x in
            self.get_mappings_guaranteed(
                source_child_id, FastMapDirection.source_to_grav, mapping_type, source_parent_id)
        ]

    def get_source_parent_ids(
        self,
        gravitate_id: str | CompositeMapKey,
        mapping_type: MappingType | None
    ) -> list[str | CompositeMapKey]:
        return [
            x.source_id for x in
            self.get_mappings_guaranteed(
                gravitate_id, FastMapDirection.grav_to_source, mapping_type)
        ]

    def get_source_child_ids(
        self,
        gravitate_parent_id: str,
        gravitate_child_id: str,
        mapping_type: MappingType | None
    ) -> list[str]:
        """
        Get all source_ids for child mappings matching the criteria.

        Note: Child lookups use string IDs only. Children mappings don't support composite keys.
        """
        return [
            x.source_id for x in
            self.get_mappings_guaranteed(
                gravitate_child_id, FastMapDirection.grav_to_source, mapping_type, gravitate_parent_id)
        ]

    def get_gravitate_parent_id_str(
        self,
        source_id: str,
        mapping_type: MappingType | None
    ) -> str:
        """Get gravitate_id as string. Raises TypeError if mapping uses composite key."""
        result = self.get_gravitate_parent_id(source_id, mapping_type)
        if isinstance(result, CompositeMapKey):
            raise TypeError(f"Mapping has composite gravitate_id: {result}")
        return result

    def get_source_parent_id_str(
        self,
        gravitate_id: str,
        mapping_type: MappingType | None
    ) -> str:
        """Get source_id as string. Raises TypeError if mapping uses composite key."""
        result = self.get_source_parent_id(gravitate_id, mapping_type)
        if isinstance(result, CompositeMapKey):
            raise TypeError(f"Mapping has composite source_id: {result}")
        return result

    # Composite key lookup methods

    def get_mappings_by_composite(
        self,
        from_key: Union[dict, CompositeMapKey],
        direction: FastMapDirection,
        mapping_type: MappingType | None,
        parent_from_id: Optional[str] = None
    ) -> list[Map | Children]:
        """
        Retrieve mappings using a composite key for entity identity.

        :param from_key: Dict or CompositeMapKey for lookup (entity identity)
        :param direction: Mapping direction
        :param mapping_type: Type of mapping
        :param parent_from_id: Parent identifier (string only - hierarchy uses simple IDs)

        Note: Composite keys are for entity identity, not hierarchy. If looking up children,
        use string parent_from_id. Children mappings don't support composite keys.
        """
        self._raise_if_not_loaded()
        if isinstance(from_key, dict):
            from_key = CompositeMapKey(key=from_key)
        return self.fast_maps.get(
            FastMapKey(from_key, mapping_type, parent_from_id, direction), []
        )

    def get_mapping_by_composite(
            self,
            from_key: Union[dict, CompositeMapKey],
            direction: FastMapDirection,
            mapping_type: MappingType | None
    ) -> Map | Children:
        """Get single mapping by composite key, raises if not found or multiple."""
        mappings = self.get_mappings_by_composite(from_key, direction, mapping_type)
        if isinstance(from_key, dict):
            from_key = CompositeMapKey(key=from_key)
        if not mappings:
            raise RitaMapperKeyError(f"No mapping found for composite key: {from_key}", from_key=from_key)
        if len(mappings) > 1:
            raise RitaMapperKeyError(f"Multiple mappings ({len(mappings)}) found for: {from_key}", from_key=from_key)
        return mappings[0]

    def get_gravitate_id_by_composite(
        self,
        source_key: Union[dict, CompositeMapKey],
        mapping_type: MappingType | None
    ) -> str | CompositeMapKey:
        """Get gravitate_id using a composite source key."""
        return self.get_mapping_by_composite(source_key, FastMapDirection.source_to_grav, mapping_type).gravitate_id

    def get_source_id_by_composite(
        self,
        gravitate_key: Union[dict, CompositeMapKey],
        mapping_type: MappingType | None
    ) -> str | CompositeMapKey:
        """Get source_id using a composite gravitate key."""
        return self.get_mapping_by_composite(gravitate_key, FastMapDirection.grav_to_source, mapping_type).source_id

    def find_mappings_by_partial_key(
        self,
        partial_key: dict[str, str],
        direction: FastMapDirection,
        mapping_type: MappingType | None = None
    ) -> list[Map | Children]:
        """
        Find all mappings where the composite key contains the partial key fields.

        Example: partial_key={"product": "X"} matches {"product": "X", "terminal": "Y"}
        """
        self._raise_if_not_loaded()
        results = []
        for fast_key, mappings in self.fast_maps.items():
            if fast_key.direction != direction:
                continue
            if mapping_type is not None and fast_key.type != mapping_type:
                continue
            # Check if from_id is a CompositeMapKey and matches partial
            if isinstance(fast_key.from_id, CompositeMapKey):
                if fast_key.from_id.matches(partial_key):
                    results.extend(mappings)
        return results


class RitaMapper(RitaMapperCore):
    def __init__(self, provider: MappingProvider | AsyncMappingProvider, source_system: str):
        super().__init__()
        self.provider = provider
        self.source_system = source_system

    async def load_mappings_async(self):
        """
        Load mappings from the provider asynchronously.

        :raises TypeError: If the provider is not async.

        Note that this method does not check whether mappings are already loaded; calling it will overwrite any stored
        mappings with a fresh set from the provider.
        """
        if isinstance(self.provider, AsyncMappingProvider):
            mappings = await self.provider.get_mappings_by_source_system(self.source_system)
        else:
            raise TypeError("Provider is not async")
        super()._load_mappings(mappings)

    def load_mappings(self):
        """
        Load mappings from the provider synchronously.

        :raises TypeError: If the provider is async.

        Note that this method does not check whether mappings are already loaded; calling it will overwrite any stored
        mappings with a fresh set from the provider.
        """
        if isinstance(self.provider, MappingProvider):
            mappings = self.provider.get_mappings_by_source_system(self.source_system)
        else:
            raise TypeError("Provider is not synchronous")
        super()._load_mappings(mappings)



import enum
from datetime import datetime, UTC, timedelta
from typing import Any, List

import pandas_gbq
from google.oauth2 import service_account
from loguru import logger
import pandas as pd
from pymongo import MongoClient
from pymongo.synchronous.database import Database

from bb_integrations_lib.models.pipeline_structs import StopPipeline, StopBranch
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.util.utils import init_db

class OrderType(enum.Enum):
    ANY = enum.auto()
    NOT_BACKHAUL = enum.auto()
    BACKHAUL_ONLY = enum.auto()

    def to_mongo_query(self):
        match self:
            case OrderType.ANY:
                return {}
            case OrderType.NOT_BACKHAUL:
                return {"type": {"$ne": "backhaul"}}
            case OrderType.BACKHAUL_ONLY:
                return {"type": "backhaul"}

    def __str__(self) -> str:
        return {
            OrderType.ANY: "any",
            OrderType.NOT_BACKHAUL: "not backhaul",
            OrderType.BACKHAUL_ONLY: "backhaul only",
        }[self]

class GetOrderNumbersToExportStep(Step):
    def __init__(self, mongo_database: Database, order_type_filter: OrderType, exported_order_table_name: str,
                 project_id: str, gcp_credentials_file: str, testing_date_min: datetime | None = None,
                 testing_date_max: datetime | None = None, lookback_days: int = 60, use_old_change_query: bool = False,
                 check_for_updated_orders: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mongo_database = mongo_database
        self.order_type_filter = order_type_filter
        self.exported_order_table_name = exported_order_table_name
        self.project_id = project_id
        self.gcp_credentials_file = gcp_credentials_file
        self.testing_date_min = testing_date_min
        self.testing_date_max = testing_date_max
        self.lookback_days = lookback_days
        self.use_old_change_query = use_old_change_query
        self.check_for_updated_orders = check_for_updated_orders
        if not self.gcp_credentials_file.endswith(".json"):
            self.gcp_credentials_file += ".json"

    def describe(self) -> str:
        return "Determine which order numbers to export."

    async def execute(self, i: Any) -> List[int]:
        collection = self.mongo_database["order_v2"]
        if not self.testing_date_min and not self.testing_date_max:
            lookback_date = datetime.now(UTC) - timedelta(days=self.lookback_days)
            date_query = {'$or': [
                {'updated_on': {'$gte': lookback_date}},
                {'movement_updated': {'$gte': lookback_date}}
            ]}
            logger.debug(f"Looking for orders in the previous {self.lookback_days} days.")
        elif self.testing_date_min and self.testing_date_max:
            date_query = {'$or': [
                {'updated_on': {'$gte': self.testing_date_min, '$lte': self.testing_date_max}},
                {'movement_updated': {'$gte': self.testing_date_min, '$lte': self.testing_date_max}}
            ]}
            logger.debug(f"Looking for orders between {self.testing_date_min} and {self.testing_date_max}")
        else:
            raise RuntimeError("testing_date_max and testing_date_min must provided together, or both not provided.")

        if self.use_old_change_query:
            change_query = {"change_date": "$updated_on"}
        else:
            change_query = {"change_date": { "$max": ["$updated_on", "$movement_updated"] }}

        # Over time the number of orders will grow. To keep step performance constant we're only going to look at
        # orders updated within the last 60 days
        assert isinstance(self.order_type_filter, OrderType)
        logger.debug(f"Searching for orders of type: {str(self.order_type_filter)}")
        orders = list(collection.find({**date_query,
                                       **self.order_type_filter.to_mongo_query(),
                                       "state": "complete"},
                                      {
                                          "_id": 0,
                                          "number": 1,
                                          **change_query,
                                      }))
        logger.debug(f"There are {len(orders)} orders updated in the last {self.lookback_days} days.")
        if len(orders) == 0:
            raise StopBranch()
        # We need to look up the list of exported orders from GBQ
        credentials = service_account.Credentials.from_service_account_file(self.gcp_credentials_file)
        sql = (f"select order_number as number, max(date) as exported_date"
               f" from `{self.exported_order_table_name}`"
               f" group by order_number;")
        gbq_df = pandas_gbq.read_gbq(sql, project_id=self.project_id, credentials=credentials, progress_bar_type=None)
        logger.debug(f"Received {gbq_df.shape[0]} GBQ records.")

        # Create a dataframe from the list of orders and do a left join on the orders exported from GBQ
        orders_df = pd.DataFrame.from_records(orders)
        joined = pd.merge(orders_df, gbq_df, on="number", how="left")

        never_exported = joined[joined["exported_date"].isna()]
        logger.debug(f"There are {never_exported.shape[0]} new orders to export.")

        if self.check_for_updated_orders:
            updated_after_export = joined[joined["exported_date"] < joined["change_date"]]
            logger.debug(f"There are {updated_after_export.shape[0]} orders that have been updated and need to be exported again.")
            to_export = set(never_exported["number"]).union(set(updated_after_export["number"]))
        else:
            to_export = set(never_exported["number"])

        output = list(to_export)
        logger.debug(output)
        if len(output) == 0:
            # There are no orders to export. Cancel the pipeline this step is part of.
            raise StopBranch()
        return output


if __name__ == "__main__":
    import asyncio
    async def main():
        s = GetOrderNumbersToExportStep(
            mongo_database=MongoClient("mongodb conn str")["db_name"],
            order_type_filter=OrderType.ANY,
            exported_order_table_name="gravitate-harms-prod.harms_order_exports.exported_backhaul_orders",
            project_id="gravitate-harms-prod",
            gcp_credentials_file="google.credentials.json"
        )
        await s.execute(None)
    asyncio.run(main())



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class DriverExtraData(BaseModel):
    source_id: Optional[str] = None
    source_system_id: Optional[str] = None


class Drop(BaseModel):
    compartment_indexes: list[int]
    before_stick_inches: Optional[int] = None
    before_stick_time: Optional[datetime] = None
    after_stick_inches: Optional[int] = None
    after_stick_time: Optional[datetime] = None
    drop_id: str
    location: str
    location_id: str
    location_source_id: Optional[str] = None
    location_source_system: Optional[str] = None
    product: str
    product_id: str
    product_source_id: Optional[str] = None
    product_source_system: Optional[str] = None
    tank_id: Optional[int] = None
    tank_source_id: Optional[str] = None
    tank_source_system: Optional[str] = None
    volume: int
    driver_id: Optional[str] = None
    driver_username: Optional[str] = None
    driver_extra_data: Optional[DriverExtraData] = None
    driver_name: Optional[str] = None
    driver_date: Optional[datetime] = None
    driver_shift: Optional[str] = None
    driver_shift_id: Optional[str] = None
    tractor_number: Optional[str] = None
    trailer_number: Optional[str] = None
    blend_codes: list[str] = []
    site_counterparty_id: Optional[str] = None
    site_counterparty_name: Optional[str] = None
    freight_customer_id: Optional[str] = None
    freight_customer_name: Optional[str] = None


class BOLPriceDetails(BaseModel):
    price: Optional[float] = None
    price_id: Optional[str] = None
    curve_id: Optional[str] = None
    date: Optional[datetime] = None
    counterparty_id: Optional[str] = None


class BOLDetail(BaseModel):
    compartment_index: Optional[int] = None
    load_number: Optional[str] = None
    product: str
    product_id: str
    product_source_id: Optional[str] = None
    product_source_system: Optional[str] = None
    net_volume: int
    gross_volume: int
    supplier: Optional[str] = None
    supplier_id: Optional[str] = None
    supplier_source_id: Optional[str] = None
    supplier_source_system: Optional[str] = None
    contract: Optional[str] = None
    price_details: Optional[BOLPriceDetails] = None


class BOL(BaseModel):
    bol_number: str
    bol_id: str
    date: Optional[datetime] = None
    details: list[BOLDetail] = []
    load_or_drop: Optional[str] = None
    location: Optional[str] = None
    location_id: Optional[str] = None
    location_source_id: Optional[str] = None
    location_source_system: Optional[str] = None
    driver_id: Optional[str] = None
    driver_username: Optional[str] = None
    driver_extra_data: Optional[DriverExtraData] = None
    driver_name: Optional[str] = None
    driver_date: Optional[datetime] = None
    driver_shift: Optional[str] = None
    driver_shift_id: Optional[str] = None
    tractor_number: Optional[str] = None
    trailer_number: Optional[str] = None


class Cost(BaseModel):
    cost_type: str
    carrier: Optional[str] = None
    carrier_source_id: Optional[str] = None
    carrier_source_system: Optional[str] = None
    per_unit_price: Optional[float] = None
    cost_amount: Optional[float] = None


class FreightLeg(BaseModel):
    origin: str
    origin_id: str
    destination: str
    destination_id: str
    distance: Optional[float] = None


class FreightTransaction(BaseModel):
    model_type: Optional[str] = None
    type: Optional[str] = None
    subtype: Optional[str] = None
    rate: Optional[float] = None
    amount: Optional[float] = None
    total: Optional[float] = None
    uom: Optional[str] = None
    errors: list[str] = []
    id: str
    product_group: Optional[str] = None
    product_id: Optional[str] = None
    product_name: Optional[str] = None
    origin: Optional[str] = None
    origin_id: Optional[str] = None
    origin_override: Optional[str] = None
    origin_id_override: Optional[str] = None
    destination: Optional[str] = None
    destination_id: Optional[str] = None
    destination_override: Optional[str] = None
    destination_id_override: Optional[str] = None
    legs: list[FreightLeg] = []
    use_surcharge: Optional[bool] = None
    manual: Optional[bool] = None
    has_dependency: Optional[bool] = None
    created_date: Optional[datetime] = None
    requires_approval: Optional[bool] = None
    requires_approval_reason_code: Optional[bool] = None
    is_approved: Optional[bool] = None
    approved_by: Optional[str] = None
    approved_date: Optional[datetime] = None
    approved_reason_code: Optional[str] = None
    amount_override: Optional[float] = None
    rate_override: Optional[float] = None
    product_id_override: Optional[str] = None
    product_name_override: Optional[str] = None
    product_group_override: Optional[str] = None
    subtype_override: Optional[str] = None
    uom_override: Optional[str] = None
    total_override: Optional[float] = None
    override_by: Optional[str] = None
    override_date: Optional[datetime] = None
    exclude_from_invoice: Optional[bool] = None
    credit_rebill_metadata: Optional[dict] = None
    extra_data: Optional[dict] = None
    gross_volume: Optional[int] = None
    net_volume: Optional[int] = None
    bol_date: Optional[datetime] = None
    bol_number: Optional[str] = None
    bol_number_override: Optional[str] = None
    delivery_date: Optional[datetime] = None


class Freight(BaseModel):
    id: str
    parent: Optional[str] = None
    created_at: Optional[datetime] = None
    type: Optional[str] = None
    status: Optional[str] = None
    reversed: Optional[bool] = None
    transactions: list[FreightTransaction] = []
    counterparty_id: Optional[str] = None
    book_type: Optional[str] = None
    effective_date_used: Optional[datetime] = None
    invoice_group_id: Optional[str] = None
    invoice_number: Optional[str] = None
    accessorial_invoice_number: Optional[str] = None
    new_version_available: Optional[bool] = None
    new_version_total: Optional[float] = None
    error: Optional[str] = None


class AllocatedBOLPrice(BaseModel):
    contract: Optional[str] = None
    price_type: Optional[str] = None
    price_id: Optional[str] = None
    curve_id: Optional[str] = None
    price: Optional[float] = None
    date: Optional[datetime] = None
    expired: Optional[bool] = None


class AllocatedBOL(BaseModel):
    bol_number: str
    bol_id: str
    bol_terminal_id: Optional[str] = None
    bol_terminal: Optional[str] = None
    bol_product: Optional[str] = None
    bol_product_id: Optional[str] = None
    bol_supplier: Optional[str] = None
    bol_supplier_id: Optional[str] = None
    bol_net_volume_original: Optional[int] = None
    bol_gross_volume_original: Optional[int] = None
    store_number: Optional[str] = None
    store_id: Optional[str] = None
    location_id: Optional[str] = None
    store_tank: Optional[int] = None
    store_product: Optional[str] = None
    store_product_id: Optional[str] = None
    store_timezone: Optional[str] = None
    bol_net_volume_allocated: Optional[int] = None
    bol_gross_volume_allocated: Optional[int] = None
    driver_schedule: Optional[str] = None
    bol_date: Optional[datetime] = None
    delivered_date: Optional[datetime] = None
    price: Optional[AllocatedBOLPrice] = None


class Order(BaseModel):
    order_number: str
    order_id: str
    po: Optional[str] = None
    carrier: Optional[str] = None
    carrier_id: Optional[str] = None
    manager: Optional[str] = None
    manager_id: Optional[str] = None
    last_movement_update: Optional[datetime] = None
    order_date: Optional[datetime] = None
    status: Optional[str] = None
    type: Optional[str] = None
    drops: list[Drop] = []
    bols: list[BOL] = []
    costs: list[Cost] = []
    validation_bypass_on: Optional[datetime] = None
    has_additives: Optional[bool] = None
    estimated_freight: Optional[Freight] = None
    actual_freight: Optional[Freight] = None
    allocated_bol_error: Optional[str] = None
    allocated_bol_issue: Optional[str] = None
    allocated_bols: list[AllocatedBOL] = []
    last_change_date: Optional[datetime] = None
    reference_order_number: Optional[str] = None